import pandas as pd
import numpy as np
from sklearn.decomposition import PCA
from itertools import chain


# filter out mitochondrial and ribosomal genes
def filter_out_non_coding_genes(adata):

    # remove the mitochondrial and ribosomal genes
    adata.var["mt"] = adata.var_names.str.startswith("MT-")
    adata.var["ribosomal"] = adata.var_names.str.startswith(("RPS", "RPL"))

    adata = adata[:, adata.var['mt'] == False]
    adata = adata[:, adata.var['ribosomal'] == False]

    # remove genes that have a dot in their name
    adata = adata[:, ~adata.var_names.str.contains("\.")]

    # remove genes whose name starts with "Gm" or "GM"
    adata = adata[:, ~adata.var_names.str.startswith(("Gm", "GM"))]

    # remove genes whose last letter is "y" or "Y"
    adata = adata[:, ~adata.var_names.str.endswith(("y", "Y"))]

    # remove genes whose first 3 letters are HLA or Hla
    adata = adata[:, ~adata.var_names.str.startswith(("HLA", "Hla"))]

    # remove "xis" genes
    adata = adata[:, ~adata.var_names.str.startswith(("Xis", "XIS", "xis"))]

    # remove genes whose first 3 letters are "TSI" or "Tsi"
    adata = adata[:, ~adata.var_names.str.startswith(("TSI", "Tsi"))]

    # keep only the protein coding genes
    adata = adata[:, adata.var['protein_coding'] == True]

    return adata


# Variance based feature selection
def variance_based_feature_selection(adata, fold_threshold=0.5, samples_differing=3):

    adata.var['udon_hvg'] = True
    pseudobulk_folds_df = adata.varm['pseudobulk_folds']

    for gene in adata.var_names:
        gene_vector = pseudobulk_folds_df.loc[gene, :]
        gene_vector = gene_vector.sort_values(ascending=True)

        n_lowest = gene_vector.head(samples_differing).iloc[samples_differing-1]
        n_highest = gene_vector.tail(samples_differing).iloc[0]
        fold = n_highest - n_lowest
        adata.var.loc[gene, 'udon_hvg'] = fold > fold_threshold

    return adata


# Inter correlation step do this step if only less than 8000 to start with -- cut off for this 0.4 -- should bring it down to 1500
def intercorrelation_based_feature_selection(adata, corr_threshold=0.4, corr_n_events=5):

    adata_og = adata.copy()
    adata = adata[:, adata.var['udon_hvg'] == True]

    pseudobulk_folds_df = adata.varm['pseudobulk_folds']
    # Calculate correlation matrix
    pseudobulk_folds_df = pseudobulk_folds_df.transpose()

    corr_matrix = pseudobulk_folds_df.corr(method="pearson")

    # Get columns with correlation > 0.4
    correlated_features = (corr_matrix > corr_threshold).sum() > corr_n_events

    # Update adata.var with correlated features
    adata_og.var['correlated_genes'] = False
    adata_og.var.loc[correlated_features.index[correlated_features], 'correlated_genes'] = True

    return adata_og


# pca based variable feature selection
def pca_feature_selection(adata, corr_threshold=0.4, n_components=30):

    adata_og = adata.copy()
    adata = adata[:, adata.var['correlated_genes'] == True]

    exp = adata.varm['pseudobulk_folds']
    pca_obj = PCA(n_components, svd_solver='full')

    pca_obj.fit_transform(exp.transpose())
    loadings = pd.DataFrame(pca_obj.components_.T, index=exp.index)
    correlated_features_all_pc = []

    for pc in range(1, n_components + 1):
        # print(pc)
        pc_i = pd.Series(loadings.loc[:, pc - 1])  # get the genes and their loadings value from the given PC
        pc_i_sorted = pc_i.sort_values(ascending=False)

        pc_top200 = pc_i_sorted.head(200).index.tolist()  # top 200 features
        pc_bottom200 = pc_i_sorted.tail(200).index.tolist()  # bottom 200 features

        pc_400features = pc_top200 + pc_bottom200

        exp_400_features_pc = exp.loc[pc_400features, :]  # limit the expression to the above calculated 400 genes for that PC

        cor_400features_matrix = exp_400_features_pc.T.corr(method="pearson")
        cor_400features_matrix = np.tril(cor_400features_matrix, k=-1)
        idx_cor_400features_binary = abs(cor_400features_matrix) > corr_threshold
        idx_cor_400features_binary = idx_cor_400features_binary.astype(int)
        idx_cor_400features_binary = pd.DataFrame(idx_cor_400features_binary)
        idx_cor_400features_binary.index = exp_400_features_pc.index.tolist()
        idx_cor_400features_binary.columns = exp_400_features_pc.index.tolist()

        z = idx_cor_400features_binary.sum(axis=0)
        z = z[z > 0].index.tolist()
        z2 = idx_cor_400features_binary.sum(axis=1)
        z2 = z2[z2 > 0].index.tolist()

        correlated_events_pc = list(set(z) & set(z2))
        correlated_features_all_pc.append(correlated_events_pc)

    correlated_events_all_pc_unique = list(chain.from_iterable(correlated_features_all_pc))
    correlated_events_all_pc_unique = list(set(correlated_events_all_pc_unique))

    # Update adata.var with pca_selected_genes
    adata_og.var['pca_selected_genes'] = False
    adata_og.var.loc[correlated_events_all_pc_unique, 'pca_selected_genes'] = True

    return adata_og


def feature_selection_wrapper(adata, species, gtf_file_path=None, fold_threshold=1, samples_differing=3, intercorr_threshold=0.4, corr_n_events=5, pca_corr_threshold=0.4, n_components=30):

    # identify_protein_coding_genes
    print("Identifying protein coding genes...")
    adata = identify_protein_coding_genes(adata, species=species, gtf_file_path=gtf_file_path)

    print("before protein coding genes")
    print(adata.shape[1])
    # filter_out_non_coding_genes
    print("Filtering out non-coding genes...")
    adata = filter_out_non_coding_genes(adata)
    print("after protein coding genes")
    print(adata.shape[1])

    # variance_based_feature_selection
    print("Performing variance-based feature selection...")
    adata = variance_based_feature_selection(adata, fold_threshold=fold_threshold, samples_differing=samples_differing)

    print("after variance based feature selection")
    # get length of variables for which adata.var['udon_hvg'] is True
    print(adata.var['udon_hvg'].sum())

    # intercorrelation_based_feature_selection
    print("Performing intercorrelation-based feature selection...")
    adata = intercorrelation_based_feature_selection(adata, corr_threshold=intercorr_threshold, corr_n_events=corr_n_events)
    # filter the adata for only the correlated features/genes
    print("after intercorrelation based feature selection")
    # get length of variables for which adata.var['correlated_genes'] is True
    print(adata.var['correlated_genes'].sum())

    # pca_feature_selection
    print("Performing PCA-based feature selection...")
    adata = pca_feature_selection(adata, corr_threshold=pca_corr_threshold, n_components=n_components)
    print("after pca feature selection")
    # get length of variables for which adata.var['pca_selected_genes'] is True
    print(adata.var['pca_selected_genes'].sum())

    print("feature selection completed!")

    return adata


def identify_protein_coding_genes(adata, species, gtf_file_path=None):

    # read gtf file
    if gtf_file_path is not None:
        # read gtf file
        gtf_df = read_gtf_file(gtf_file_path)

        # Filter protein-coding genes
        protein_coding_genes = gtf_df[
            (gtf_df['feature'] == 'gene') & (gtf_df['attributes'].str.contains('protein_coding'))]
        protein_coding_gene_names = protein_coding_genes['attributes'].str.extract(r'gene_name "(.+?)";',
                                                                          expand=False).dropna().unique()

    else:
        gtf_df = pd.read_csv('ProteinCoding-Hs-Mm.txt', sep="\t", names=["GeneName"])
        # Filter based on species
        if species == "Hs":
            # Filter uppercase entries or those starting with "ENSG"
            gtf_df = gtf_df[(gtf_df['GeneName'].str.isupper()) | (gtf_df['GeneName'].str.startswith('ENSG'))]
        elif species == "Mm":
            # Filter uppercase entries or those starting with "ENSG"
            gtf_df = gtf_df[~(gtf_df['GeneName'].str.isupper()) | (gtf_df['GeneName'].str.startswith('ENSMU'))]

        protein_coding_gene_names = gtf_df['GeneName'].unique()

    gene_list = adata.var_names
    protein_coding_genes_in_list = [gene for gene in gene_list if gene.split("_")[0] in protein_coding_gene_names]

    # label protein_coding_genes_in_list as "protein_coding" in adata.var
    adata.var['protein_coding'] = False
    adata.var.loc[protein_coding_genes_in_list, 'protein_coding'] = True

    return adata


def read_gtf_file(gtf_file_path):

    columns = ["seqname", "source", "feature", "start", "end", "score", "strand", "frame", "attributes"]
    df = pd.read_csv(gtf_file_path, sep='\t', comment='#', names=columns, dtype={"seqname": str})

    return df



import pandas as pd
import numpy as np
import os,sys
import re
from tqdm import tqdm
import argparse
import multiprocessing as mp


def build_gene_model(df):
    # column names are: chr, source, type, start, end, score, strand, phase, attrs
    transcript_id = -1
    position_footprint = {}
    position_ense = {}
    gene_start, gene_end, strand, ensg, chromsome = None,None,None,None,None
    pat = re.compile(r'exon_id "(ENSE\d+)"')
    # phase1: build footprint
    for row in df.itertuples(index=False):
        if row.type == 'gene':
            gene_start, gene_end, strand, ensg, chromosome = row.start, row.end, row.strand, row.gene, row.chr
        elif row.type == 'transcript':
            transcript_id += 1
            continue
        elif row.type == 'exon':
            attrs = row.attrs
            ense = re.search(pat,attrs).group(1)
            for p in range(row.start,row.end+1,1):
                # put into position_footprint
                position_footprint.setdefault(p,[]).append(transcript_id)
                # put into position_ense
                position_ense.setdefault(p,[]).append(ense)
    # phase2: moving along the footprint
    if strand == '+':
        string_stream = ''
        running_profile = position_footprint[gene_start]
        block_index = 1
        intron_block_index = 0
        segment_index = 1
        anchor_position = gene_start
        p = gene_start
        while p <= gene_end:
            try:
                position_footprint[p]
            except KeyError:
                # write
                subexon_identifier = 'E'+str(block_index)+'.'+str(segment_index)
                associated_ense = '|'.join(position_ense[p-1])
                string_stream += '{}\t{}\t{}\t{}\t{}\t{}\t{}\n'.format(ensg,chromosome,strand,subexon_identifier,anchor_position,p-1,associated_ense)  # because current p is the first base in intron, shouldn't be included in previous exon
                # enumerate the whole intron region
                pi = p
                while True:
                    try:
                        position_footprint[pi]
                    except KeyError:
                        pi += 1
                    else:
                        # write the intron
                        intron_block_index += 1
                        subexon_identifier = 'I'+str(intron_block_index)+'.'+str(1)
                        string_stream += '{}\t{}\t{}\t{}\t{}\t{}\t{}\n'.format(ensg,chromosome,strand,subexon_identifier,p,pi-1,'')  # because the current pi is the first base in next exon, shouldn't be included in the intron
                        # update
                        p = pi
                        running_profile = position_footprint[p]
                        block_index += 1
                        segment_index = 1
                        anchor_position = p
                        break
            else:
                if position_footprint[p] != running_profile:   
                    # write          
                    subexon_identifier = 'E'+str(block_index)+'.'+str(segment_index)
                    associated_ense = '|'.join(position_ense[p-1])  # because the current p is the first base in next segment, we are writing the previous segment and its associated ense
                    string_stream += '{}\t{}\t{}\t{}\t{}\t{}\t{}\n'.format(ensg,chromosome,strand,subexon_identifier,anchor_position,p-1,associated_ense)
                    # update
                    running_profile = position_footprint[p]
                    segment_index += 1
                    anchor_position = p
                    p += 1
                else:
                    p += 1
        # write the last exon segment
        subexon_identifier = 'E'+str(block_index)+'.'+str(segment_index)
        associated_ense = '|'.join(position_ense[p-1])
        string_stream += '{}\t{}\t{}\t{}\t{}\t{}\t{}\n'.format(ensg,chromosome,strand,subexon_identifier,anchor_position,p-1,associated_ense)

    else:   # negative string, so backtrack
        string_stream = ''
        running_profile = position_footprint[gene_end]
        block_index = 1
        intron_block_index = 0
        segment_index = 1
        anchor_position = gene_end
        p = gene_end
        while p >= gene_start:
            try:
                position_footprint[p]
            except KeyError:
                # write
                subexon_identifier = 'E'+str(block_index)+'.'+str(segment_index)
                associated_ense = '|'.join(position_ense[p+1])
                string_stream += '{}\t{}\t{}\t{}\t{}\t{}\t{}\n'.format(ensg,chromosome,strand,subexon_identifier,p+1,anchor_position,associated_ense)  
                # enumerate the whole intron region
                pi = p
                while True:
                    try:
                        position_footprint[pi]
                    except KeyError:
                        pi -= 1
                    else:
                        # write the intron
                        intron_block_index += 1
                        subexon_identifier = 'I'+str(intron_block_index)+'.'+str(1)
                        string_stream += '{}\t{}\t{}\t{}\t{}\t{}\t{}\n'.format(ensg,chromosome,strand,subexon_identifier,pi+1,p,'')  
                        # update
                        p = pi
                        running_profile = position_footprint[p]
                        block_index += 1
                        segment_index = 1
                        anchor_position = p
                        break
            else:
                if position_footprint[p] != running_profile:   
                    # write          
                    subexon_identifier = 'E'+str(block_index)+'.'+str(segment_index)
                    associated_ense = '|'.join(position_ense[p+1])  
                    string_stream += '{}\t{}\t{}\t{}\t{}\t{}\t{}\n'.format(ensg,chromosome,strand,subexon_identifier,p+1,anchor_position,associated_ense)
                    # update
                    running_profile = position_footprint[p]
                    segment_index += 1
                    anchor_position = p
                    p -= 1
                else:
                    p -= 1
        # write the last exon segment
        subexon_identifier = 'E'+str(block_index)+'.'+str(segment_index)
        associated_ense = '|'.join(position_ense[p+1])
        string_stream += '{}\t{}\t{}\t{}\t{}\t{}\t{}\n'.format(ensg,chromosome,strand,subexon_identifier,p+1,anchor_position,associated_ense)        
    return string_stream

def split_array_to_chunks(array,cores=None):
    if not isinstance(array,list):
        raise Exception('split_array_to_chunks function works for list, not ndarray')
    array_index = np.arange(len(array))
    if cores is None:
        cores = mp.cpu_count()
    sub_indices = np.array_split(array_index,cores)
    sub_arrays = []
    for sub_index in sub_indices:
        item_in_group = []
        for i in sub_index:
            item_in_group.append(array[i])
        sub_arrays.append(item_in_group)
    return sub_arrays

def process_single_core(chunk):
    string_stream = ''
    for sub_df in tqdm(chunk,total=len(chunk)):
        sub_string_stream = build_gene_model(sub_df)
        string_stream += sub_string_stream
    return string_stream


def main(args):
    gtf = args.gtf
    gene = args.gene
    outdir = args.outdir
    df = pd.read_csv(gtf,sep='\t',header=None,skiprows=[0,1,2,3,4])
    df.columns = ['chr','source','type','start','end','score','strand','phase','attrs']
    df = df.loc[df['type'].isin(['gene','transcript','exon']),:]
    pat = re.compile(r'gene_id "(ENSG\d+)"')
    df['gene'] = [re.search(pat,item).group(1) for item in df['attrs']]
    if gene != 'all':
        df = df.loc[df['gene']==gene,:]
        string_stream = build_gene_model(df)
    else:
        string_stream = ''
        sub_df_list = list(list(zip(*list(df.groupby(by='gene'))))[1])
        n_cores = mp.cpu_count()
        chunks = split_array_to_chunks(sub_df_list,n_cores)
        pool = mp.Pool(processes=n_cores)
        print('spawn {} subprocesses'.format(n_cores))
        r = [pool.apply_async(func=process_single_core,args=(chunk,)) for chunk in chunks]
        pool.close()
        pool.join()
        for collect in r:
            sub_string_stream = collect.get()
            string_stream += sub_string_stream

    with open(os.path.join(outdir,'gene_model_{}.txt'.format(gene)),'w') as f:
        f.write(string_stream)
        
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Build Gene Model from GTF')
    parser.add_argument('--gtf',type=str,default=None,help='the path to the gtf file')
    parser.add_argument('--gene',type=str,default=None,help='either all or stable ENSG ID')
    parser.add_argument('--outdir',type=str,default=None,help='output dir for the gene model txt file')
    args = parser.parse_args()
    main(args)


   

            








            
        










import pandas as pd
import numpy as np
from sklearn.svm import LinearSVC
from tqdm import tqdm


def generate_train_data(fold_matrix_with_mf_genes, groups):
    # for each of the nmf clusters provided, compute the centroid for each cluster for the differential splicing events using the output_file arg (provided in the output file?)

    # Set the cell types and donors for whom to create pseudobulks
    selected_cell_types = pd.unique(groups['cluster'])

    # Initialize an empty dictionary to store results
    pseudobulk_dict = {}

    # Iterate over selected cell types
    for cell_type in tqdm(selected_cell_types, desc="generating train data using centroids"):
        # Filter cells based on cell type
        selected_cells = groups.index[groups['cluster'] == cell_type]

        # Calculate average gene expression for selected cells
        avg_expression = fold_matrix_with_mf_genes.loc[:, selected_cells].mean(axis=1)

        # Add average expression values to the result dictionary
        column_name = f"U{cell_type}"
        pseudobulk_dict[column_name] = avg_expression

    # Convert the dictionary to a DataFrame
    centroids = pd.DataFrame(pseudobulk_dict)

    return centroids


def classify(train, fold_matrix_with_mf_genes, groups):

    print("using random state 42...")
    # regression fit on the train dataset
    regression = LinearSVC(random_state=42)  # alternative is OneVsRestClassifier function in sklearn -- why not this one?
    regression.fit(train.transpose(), y=train.columns)  # train should be # of ranks by # of mf genes ("samples" as per python documentation are the ranks/clusters) and "features" are genes and groups is # of ranks/clusters by # of pseudobulks/samples

    prob = regression.decision_function(fold_matrix_with_mf_genes.transpose())  # it should be a # of samples/patients by # of features format so we are transposing it

    # Determine the cluster with the maximum probability for each sample
    final_clusters = np.argmax(prob, axis=1)
    final_clusters = pd.DataFrame(train.columns[final_clusters], index=fold_matrix_with_mf_genes.columns, columns=['cluster'])
    final_clusters = final_clusters.sort_values('cluster')

    return final_clusters




#!/usr/bin/env python
# coding: utf-8

# ##### mwu6.py
# 
# May 3, 2023
# 
# Mann-Whitney U non-parametric (rank-based) test (aka Wilcoxon for 2 independent groups)
# 


# ### Define function to perform some data checks before mwu computation

def mwuChecks( dataf, grpdf, *args, grpvar='grp',min_group_size=6 ): 
#
#   This function performs some checks before calling mwuCompute().    
#
#   dataf is a pandas data frame with samples as columns and events (features) as rows.  
#   Event names must be the row names (index), not in a column.
#   Missing values (np.nan='NaN') are allowed
#
#   grpdf is a pandas data frame with a group membership column ('grp' is the default ).
#   Sample names must be the row names (index), not in a column.
#   Aside from ordering, must be 1-1 match between dataf.columns and grpdf.index
#
#   min_group_size is minimum group size for kw test to be computed.
#   min_group_size = 6 is based on statistical rule-of-thumb for Mann-Whitney 2-group test.
#
#  Returns integer to indicate course of action:
#     0 = Problem, do not run mwuCompute
#     1 = Run mwuCompute


    import pandas as pd
    import numpy as np

    
#    print(grpvar)
#    print(type(grpvar))
    
    # --- Check for 1-1 match between sample names provided in grpdf and the data matrix

    if set(dataf.columns)!= set(grpdf.index) :
        print('Group information is inconsistent with names in data matrix')
        return( 0 )
  
        
    grpCounts = grpdf[grpvar].value_counts().to_frame()  
# returns a pandas df of counts, indexed by grp, decreasing by count
    grpCounts.columns=['grpRawN']
    nGrps = grpCounts.shape[0]
    minGrpN = min( grpCounts['grpRawN'] )
    # print('nGrps, minGrpN',nGrps,minGrpN)
    
    
    # -- Handle groups <> 2 --

    if nGrps < 2 :
        print('Number of sample groups is < 2; Mann-Whitney test not conducted')
        return( 0 )
        
    if nGrps > 2 :
        print('Number of sample groups is > 2; Mann-Whitney test not conducted')
        return( 2 )
    
    # -- Don't proceed if already know a group size is < minimum --   

    if minGrpN < min_group_size:
        print('Mann-Whitney test not conducted: A group has fewer samples than minimum group size: ',minGrpN,' < ',min_group_size)
        return( 0 )
    
    return( 1 )



# ### Define function to compute the mwu test and collect summary stats (Ns, medians)

def mwuCompute(dataf, grpdf, *args, grpvar='grp' ,min_group_size=6 ): 
    #
    #   This function performs the Mann-Whitney NP (rank) two independent groups test, aka Wilcoxon 2 grp indep test.    
    #
    #   dataf is a pandas data frame with samples as columns and events (features) as rows.  
    #   Event names must be the row names (index), not in a column.
    #   Missing values (np.nan='NaN') are allowed
    #
    #   grpdf is a pandas data frame with a group membership column.
    #   Sample names must be the row names (index), not in a column.
    #   Aside from ordering, must be 1-1 match between dataf.columns and grpdf.index
    #
    #   min_group_size is minimum group size for mwu test to be computed.
    #   6 is based on statistical rule-of-thumb for Mann-Whitney 2-group test.
    #
    #   Direction of comparison is by sorted grp levels, e.g. A vs. B, 1 vs 2.
    #   Sign of result is determined by comparing test statistic value to the null mean (not comparing group medians or means)
    #
    #   Returns pandas dataframe with columns containing group Ns (xNaN), group medians and means, difference of group means ,value of mwu test statistic, sign based on test stat, p-value and FDR-adjusted p-value.
    #   Returned df has row for each event, even if test was not computed (in this case mwu values will be NaN)
    
    import pandas as pd
    import numpy as np

    import scipy.stats

    import statsmodels.stats.multitest as smm
    
   
    groups = dataf.T.groupby(grpdf[grpvar])
#    print(groups.groups)

    grpCounts = grpdf[grpvar].value_counts().to_frame()
#    display(grpCounts)
    
    grpLevels = grpCounts.index
    grpLevels = grpLevels.tolist()
#    print(grpLevels)
    SgrpLevels = sorted(grpLevels)
#    print("SgrpLevels",SgrpLevels)

    
    g1Level = SgrpLevels[0]
    g2Level = SgrpLevels[1]
#    print('g1Level',g1Level)
#    print('g2Level',g2Level)

    g1 = groups.get_group(g1Level)
    g2 = groups.get_group(g2Level)
#    print('g1 dat',g1)
#    print('g2 dat',g2)
#   count() method counts number of non-missing entries
    eventsToRun = ((g1.count() >= min_group_size) & (g2.count() >= min_group_size)).values
#    print('eventsToRun, sum',eventsToRun,sum(eventsToRun))
#    print(g1.iloc[:,eventsToRun])
#    print(g2.iloc[:,eventsToRun])

    diffGrpMeans = g1.mean() - g2.mean()

#   Set up df for results

    result = pd.DataFrame(
    {
        ('N_' + str(g1Level)): g1.count(),
        ('N_' + str(g2Level)): g2.count(),
        ('Median_' + str(g1Level)): g1.median(),
        ('Median_' + str(g2Level)): g2.median(),
        ('Mean_' + str(g1Level)): g1.mean(),
        ('Mean_' + str(g2Level)): g2.mean(),
        ('DiffMeans_' + str(g1Level) + '_m_' + str(g2Level)): diffGrpMeans,
        'mwuStat': np.nan,
        'mwuSign': np.nan,
        'mwuPval': np.nan,
        'mwuAdjPval':np.nan
    },
        index=dataf.index )
    
#    display(result)

#   Handle situation of no events meeting min_group_size criterion

    if sum(eventsToRun) == 0:
        print('Mann-Whitney test not conducted: No events meet minimum group size criterion of ',min_group_size,'.')
        return( result)

#   Compute the mwu test statistic & p-value for eventsToRun
    
    mwu = scipy.stats.mannwhitneyu(
    g1.iloc[:,eventsToRun].values, 
    g2.iloc[:,eventsToRun].values,
    alternative="two-sided",
    method="asymptotic",
    use_continuity=True,
    nan_policy='omit'
    )  
#    print(mwu)

#   Populate stat & p-value columns    
    result.loc[eventsToRun, "mwuStat"] = mwu.statistic
    result.loc[eventsToRun, "mwuPval"] = mwu.pvalue
#    display(result)

#   Determine sign (direction) of test result based on expected mean of test stat under null
    nullMean = (result[('N_' + str(g1Level))] * result[('N_' + str(g2Level))]) / 2.0 + 0.5
#    print(nullMean)
    result["mwuSign"] = np.select(
    [
        (result["mwuStat"] < nullMean),
        (result["mwuStat"] == nullMean),
        (result["mwuStat"] > nullMean),
    ],
    [-1, 0, 1],
    default=np.nan )

#   Calculate FDR of mwuPval
#   Re-create mask in case any of the calculated p-values were NaN
    eventsForFDR = (eventsToRun & (~ result['mwuPval'].isnull().values ) )

#   Do not provide alpha cutoff value to function since want adjusted p-values, not significance calls
    FDRres = smm.fdrcorrection(result.loc[eventsForFDR, "mwuPval"].to_numpy(), 
                               method="indep",is_sorted=False)

    result.loc[eventsForFDR,'mwuAdjPval'] = FDRres[1]

#    display(result)

    return( result)

#  End of file


import os
import sys
import csv
import anndata as ad
import pandas as pd
import numpy as np
from scipy.io import mmread
from scipy.sparse import lil_matrix
import collections
import argparse
from tqdm import tqdm # progress bar

sys.path.insert(1, os.path.join(sys.path[0], '..'))
from . import isoform_matrix as iso
from . import gff_process as gff_process
    
def exportConsensusIsoformMatrix(matrix_dir, isoform_association_path, gff_source=None, barcode_clusters=None, rev=False, mtx=True, sample_id=None):
    """ 
    Decompose isoforms into exon-exon and exon-junctions counts at the single-cell and pseudobulk level
    """

    # Load the 10x matrix keyed by isoforms in the first column and cell barcode cluster annotations

    if mtx:
        adata = iso.mtx_to_adata(int_folder=matrix_dir, gene_is_index=True, feature='genes.tsv', feature_col=0, barcode='barcodes.tsv', barcode_col=0, matrix='matrix.mtx', rev=rev)
    else:
        adata = iso.tsv_to_adata(matrix_dir)    

    # Ensure obs index and column names are string type
    adata.obs.index = adata.obs.index.astype(str)
    adata.obs.columns = adata.obs.columns.astype(str)

    # If existing cell clusters are provided already (e.g., supervised classification from gene-level analyses)
    if barcode_clusters is not None and not barcode_clusters.empty:
        adata = iso.calculate_barcode_match_percentage(adata, barcode_clusters)
    else:
        if sample_id is not None:
            adata.obs['cluster'] = sample_id
        else:
            adata.obs['cluster'] = gff_source

    
    # Need to filter cell barcodes to those with a sufficient number of expressed genes
    adata.obs['total_counts'] = np.sum(adata.X, axis=1)
    adata = adata[adata.obs['total_counts'] >= 100, :]
    remaining_cells = adata.n_obs
    print(f"Number of remaining cells with >=100 reads: {remaining_cells}")

    isoform_names = adata.var.index.tolist()
    print('isoform features |',isoform_names[:5])

    def parse_isoform_mapping(isoform_association_path, gff_source = None):
        """ Parse the isoform mapping file and return dictionaries of isoform-to-collapsed-isoforms """
        isoform_to_meta_isoform = collections.OrderedDict()

        print ('Mapping isoforms to meta-isoforms...',)
        unique_isoforms={}
        unique_ref_isoforms={}
        isoform_to_gene = {}
        with open(isoform_association_path, 'r') as file:
            reader = csv.reader(file, delimiter='\t')
            for row in reader:
                #ENSG00000241860	PB.10.25	ND167_MPP_3k	ENST00000655252.1	gencode.v45.annotation
                gene, ref_isoform, ref_gff, source_isoform, source_gff = row
                if source_gff == gff_source:
                    isoform_to_meta_isoform[source_isoform] = gene+':'+ref_isoform
                    unique_isoforms[source_isoform,source_gff]=[]
                    unique_ref_isoforms[ref_isoform,ref_gff]=[]
                if ref_gff == gff_source:
                    isoform_to_meta_isoform[ref_isoform] = gene+':'+ref_isoform
                    unique_isoforms[ref_isoform,ref_gff]=[]
                    unique_ref_isoforms[ref_isoform,ref_gff]=[]
                isoform_to_gene[ref_isoform] = gene

        print('unique isoforms |',len(unique_ref_isoforms),'out of',len(unique_isoforms))
        return isoform_to_meta_isoform, isoform_to_gene

    # Load isoform mapping information
    isoform_to_meta_isoform, isoform_to_gene = parse_isoform_mapping(isoform_association_path, gff_source=gff_source)

    isoform_counts = {}
    # Iterate through each cell barcode and isoform
    print ('Computing junctions counts for cells')
    cell_data = adata.X.toarray()
    isoform_ids = adata.var_names

    for isoform_idx, isoform in tqdm(enumerate(isoform_ids), total=len(isoform_ids), desc="Processing isoforms"):
        counts = cell_data[:, isoform_idx]
        if np.any(counts > 0):
            if isoform in isoform_to_meta_isoform: # Exclude isoforms without gene annotations
                ref_isoform = isoform_to_meta_isoform[isoform]
                for cell_index, count in zip(np.where(counts > 0)[0], counts[counts > 0]):
                    if ref_isoform not in isoform_counts:
                        isoform_counts[ref_isoform] = np.zeros(len(adata.obs_names), dtype=int)
                    isoform_counts[ref_isoform][cell_index] += count

    # Convert the counts to a sparse matrix (and report progress)
    isoform_counts_df = pd.DataFrame({k: pd.Series(v, index=adata.obs_names) for k, v in tqdm(isoform_counts.items(), desc="Creating DataFrame")})
    sparse_junction_matrix = lil_matrix(isoform_counts_df.shape)
    for i, (index, row) in enumerate(tqdm(isoform_counts_df.iterrows(), total=isoform_counts_df.shape[0], desc="Converting to sparse matrix")):
        sparse_junction_matrix[i, :] = row.values
    sparse_junction_matrix = sparse_junction_matrix.tocsr()

    # Create a new AnnData object for the junction counts
    isoform_adata = ad.AnnData(X=sparse_junction_matrix, obs=adata.obs, var=pd.DataFrame(index=isoform_counts.keys()))

    # Add gene annotations
    isoform_adata.var['gene'] = [isoform_to_gene.get(isoform, '') for isoform in isoform_adata.var_names]
    
    # Some weird key's get added when excluding barcode to cluster associations
    isoform_adata.obs = isoform_adata.obs.dropna().reset_index(drop=True)


    # Write the junction counts to an h5ad file
    h5ad_dir = os.path.basename(gff_source)
    
    #isoform_adata.write_h5ad("filtered_junction.h5ad")
    isoform_adata.write_h5ad(f"{h5ad_dir.split('.g')[0]}-isoform.h5ad", compression='gzip')
    print ('h5ad exported')

    export_pseudobulk = False
    if export_pseudobulk:
        # Compute pseudo-cluster counts and write them to a file
        grouped = isoform_counts_df.groupby(adata.obs['cluster'])
        
        # Remove clusters with less than 10 cells
        filtered_summed_groups = grouped.filter(lambda x: len(x) >= 10).groupby(adata.obs['cluster']).sum()
        
        # Remove junctions with less than 10 reads total
        filtered_summed_groups = filtered_summed_groups.loc[:, filtered_summed_groups.sum(axis=0) >= 10]
        filtered_summed_groups_transposed = filtered_summed_groups.transpose()
        filtered_summed_groups_transposed.to_csv(f"{h5ad_dir.split('.g')[0]}_isocounts.txt", sep='\t')
        print ('pseudobulk cluster junction counts exported')

    return isoform_adata

if __name__ == '__main__':
    # python iso_matrix_to_junctions.py /path/to/matrix /path/to/ensembl_exon /path/to/gff --barcode_cluster /path/to/barcode_cluster

    parser = argparse.ArgumentParser(description='Process single-cell data to generate junction matrices.')
    parser.add_argument('matrix_path', type=str, help='Path to the matrix directory.')
    parser.add_argument('ensembl_exon_dir', type=str, help='Path to the Ensembl exon directory.')
    parser.add_argument('gff_source', type=str, help='Path to the GFF source file.')
    parser.add_argument('--barcode_cluster', type=str, default=None, help='Optional: Path to the barcode cluster file.')
    parser.add_argument('--reverse_complement', type=bool, default=False, help='Optional: Reverse complement the sparse matrix cell barcodes.')
    args = parser.parse_args()

    barcode_clusters = pd.read_csv(args.barcode_cluster, sep='\t', index_col=0)
    exportConsensusIsoformMatrix(args.matrix_path,isoform_association_path,gff_source=args.gff_source,barcode_clusters=barcode_clusters,rev=args.reverse_complement)




import pytest
import pathlib
from altanalyze3.utilities.parser import (
    ArgsParser
)


DATA_FOLDER = pathlib.Path(__file__).resolve().parents[1].joinpath("data")


@pytest.mark.parametrize(
    "args, control_chr",
    [
        (
            [
                "juncount",
                "--bam", "Cal27P5-1.bam",
                "--ref", "gene_model_all.tsv"
            ],
            [
                "chr1", "chr10", "chr11", "chr12", "chr13", "chr14", "chr15",
                "chr16", "chr17", "chr18", "chr19", "chr2", "chr20", "chr21",
                "chr22", "chr3", "chr4", "chr5", "chr6", "chr7", "chr8", "chr9",
                "chrM", "chrMT", "chrX", "chrY"                                   # this will still include chrM and chrMT, because we use MAIN_CRH as default
            ]
        ),
        (
            [
                "juncount",
                "--bam", "Cal27P5-1.bam",
                "--ref", "gene_model_all.tsv",
                "--chr", "chr1"
            ],
            [
                "chr1"
            ]
        ),
        (
            [
                "juncount",
                "--bam", "Cal27P5-1.bam",
                "--ref", "gene_model_all.tsv",
                "--chr", "chr1", "chr2", "chr3"
            ],
            [
                "chr1", "chr2", "chr3"
            ]
        ),
        (
            [
                "juncount",
                "--bam", "Cal27P5-1.bam",
                "--ref", "gene_model_all.tsv",
                "--chr", "1", "2", "3"
            ],
            [
                "chr1", "chr2", "chr3"
            ]
        )
    ]
)
def test_chromosome_selection(monkeypatch, args, control_chr):
    monkeypatch.chdir(DATA_FOLDER)
    args = ArgsParser(args)
    assert sorted(args.chr) == sorted(control_chr)



import pysam
import pandas
import shutil
import logging
import itertools
import multiprocessing
from functools import partial
from collections import deque
from altanalyze3.utilities.logger import setup_logger
from altanalyze3.utilities.helpers import get_tmp_suffix
from altanalyze3.utilities.constants import (
    Job,
    Annotation,
    JunctionsParams,
    IntronsParams,
    AnnotationsParams,
    AnnMatchCat
)
from altanalyze3.utilities.io import (
    get_all_ref_chr,
    get_correct_contig,
    get_indexed_bed,
    export_counts_to_anndata
)


# https://pythonspeed.com/articles/pandas-string-dtype-memory/
# https://h2oai.github.io/db-benchmark/
# https://stackoverflow.com/questions/20459536/convert-pandas-dataframe-to-sparse-numpy-matrix-directly
# https://anndata.readthedocs.io/en/latest/generated/anndata.AnnData.write.html


def get_jun_annotation_jobs(args):
    # we include only chromosomes present both in jun_coords_location and references,
    # and user provided list. Contig is always prepended with 'chr'
    return [
        Job(
            contig=contig,
            location=args.tmp.joinpath(get_tmp_suffix())
        )
        for contig in get_all_ref_chr(args.jun_coords_location, args.threads)
            if contig in get_all_ref_chr(args.ref, args.threads) and contig in args.chr
    ]


class GroupedAnnotations():

    def __init__(self):
        self.__exact_match = []
        self.__partial_match = []
        self.__distant_match = []

    def add(self, sa, ea):                                                                                           # sa - start annotation, ea - end annotation

        sa_shift = "" if sa.position == 0 else f"""_{sa.position}"""                                                 # junction start coord if not exact exon match
        ea_shift = "" if ea.position == 0 else f"""_{ea.position + 1}"""                                             # junction end coorf if not exact exon match

        if sa.match == AnnMatchCat.CLOSEST or ea.match == AnnMatchCat.CLOSEST:                                       # not enough data to make a conclusion
            pass
        elif sa.gene == ea.gene and sa.strand == ea.strand:                                                          # the same gene and strand
            if sa.match == AnnMatchCat.EXON_END and ea.match == AnnMatchCat.EXON_START:                              #    exact match
                if sa.strand == "+":
                    self.__exact_match.append((f"""{sa.gene}:{sa.exon}-{ea.exon}""", sa.strand))                     #        start - end exons order
                elif sa.strand == "-":
                    self.__exact_match.append((f"""{sa.gene}:{ea.exon}-{sa.exon}""", sa.strand))                     #        end - start exons order
                else:
                    assert(False), "Not implemented logic"
            else:                                                                                                    #    partial match
                if sa.strand == "+":
                    self.__partial_match.append((f"""{sa.gene}:{sa.exon}{sa_shift}-{ea.exon}{ea_shift}""", sa.strand))  #     start - end exons order
                elif sa.strand == "-":
                    self.__partial_match.append((f"""{sa.gene}:{ea.exon}{ea_shift}-{sa.exon}{sa_shift}""", sa.strand))  #     end - start exons order
                else:
                    assert(False), "Not implemented logic"
        else:                                                                                                        # different gene and/or strand
            self.__distant_match.append((f"""{sa.gene}:{sa.exon}{sa_shift}-{ea.gene}:{ea.exon}{ea_shift}""", "."))   #    distant match

    def best(self):
        all_annotations = self.__exact_match + self.__partial_match + self.__distant_match
        if len(all_annotations) > 0:
            return all_annotations[0]
        else:
            return ("undefined", ".")


class RefDeque():

    def __init__(self, reference_iter):
        self.__reference_iter = reference_iter
        self.__exhausted = False
        self.__deque = deque()
        self.__buffer = []
        self.__last_ref = None
        self.contig = None
        self.start = None
        self.end = None
        self.name = None
        self.strand = None
        self.gene = None
        self.exon = None
        self.type = None

    def __set_state(self, reference, safe=None):
        safe = False if safe is None else safe
        self.contig = reference.contig
        self.start = reference.start
        self.end = reference.end
        self.name = reference.name
        self.strand = reference.strand
        self.gene, self.exon = self.name.split(":")
        self.type = self.exon[0].upper()
        logging.debug(f"""===> {"safely " if safe else ""}get new reference {self.contig}:{self.start:,}-{self.end:,}, {self.name}, {self.strand}""")
        if safe:
            self.__buffer.append(reference)

    def pop_left(self, safe=None):                 # if safe is True, we will be able to restore used items
        safe = False if safe is None else safe
        if safe:
            if len(self.__buffer) == 0:            # this is the first time we called pop_left with safe=True
                self.__last_ref = (
                    self.contig,
                    self.start,
                    self.end,
                    self.name,
                    self.strand,
                    self.gene,
                    self.exon,
                    self.type
                )
        else:                                      # we called unsafe pop_left, so keepeing buffer doesn't make sense anymore
            self.__buffer = []
            self.__last_ref = None
        try:
            self.__set_state(self.__deque.popleft(), safe)
        except IndexError:
            try:
                self.__deque.append(next(self.__reference_iter))
                self.__set_state(self.__deque.popleft(), safe)
            except StopIteration:
                self.__exhausted = True

    def restore(self):
        if len(self.__buffer) > 0 and self.__last_ref is not None:        # restore only when we have what to restore
            logging.debug("Restore references deque")
            self.__buffer.reverse()
            self.__deque.extendleft(self.__buffer)
            self.contig, self.start, self.end, self.name, self.strand, self.gene, self.exon, self.type = self.__last_ref
            self.__buffer = []
            self.__last_ref = None

    def is_empty(self):
        return self.__exhausted and len(self.__deque) == 0


def get_annotation(job, query_location, references_location, threads=None):
    threads = 1 if threads is None else threads
    all_annotations = []
    with pysam.TabixFile(str(query_location), mode="r", parser=pysam.asBed(), threads=threads) as query_handler:
        with pysam.TabixFile(str(references_location), mode="r", parser=pysam.asBed(), threads=threads) as references_handler:

            reference_iter = references_handler.fetch(get_correct_contig(job.contig, references_handler))
            ref_deque = RefDeque(reference_iter)                                                            # we need it because we can't roll back pysam iterator
            ref_deque.pop_left()                                                                            # getting a new reference from the start of the left

            for current_query in query_handler.fetch(get_correct_contig(job.contig, query_handler)):        # fetches all jucntions from the specific chromosome
                logging.debug(f"""---> get new region {current_query.contig}:{current_query.start:,}-{current_query.end:,}""")
                check_position = current_query.start                                                        # we always use start
                current_annotations = []

                while not ref_deque.is_empty():
                    if check_position < ref_deque.start:
                        logging.debug(f"""{check_position:,} is behind {ref_deque.contig}:{ref_deque.start:,}-{ref_deque.end:,}, {ref_deque.name}, {ref_deque.strand}""")
                        logging.debug("Closest match")
                        current_annotations.append(Annotation(ref_deque.gene, ref_deque.exon, ref_deque.strand, check_position, AnnMatchCat.CLOSEST))
                        break
                    elif check_position > ref_deque.end:
                        logging.debug(f"{check_position:,} is after {ref_deque.contig}:{ref_deque.start:,}-{ref_deque.end:,}, {ref_deque.name}, {ref_deque.strand}")
                        ref_deque.pop_left()
                    else:
                        while not ref_deque.is_empty() and check_position >= ref_deque.start:
                            if ref_deque.type == "E" and check_position == ref_deque.start:
                                logging.debug("Exact match - exon start")
                                current_annotations.append(Annotation(ref_deque.gene, ref_deque.exon, ref_deque.strand, 0, AnnMatchCat.EXON_START))
                            elif ref_deque.type == "E" and check_position == ref_deque.end:
                                logging.debug("Exact match - exon end")
                                current_annotations.append(Annotation(ref_deque.gene, ref_deque.exon, ref_deque.strand, 0, AnnMatchCat.EXON_END))
                            elif check_position > ref_deque.start and check_position < ref_deque.end:
                                if ref_deque.type == "E":
                                    logging.debug("Not exact exon match")
                                    current_annotations.append(Annotation(ref_deque.gene, ref_deque.exon, ref_deque.strand, check_position, AnnMatchCat.EXON_MID))
                                elif ref_deque.type == "I":
                                    logging.debug("Not exact intron match")
                                    current_annotations.append(Annotation(ref_deque.gene, ref_deque.exon, ref_deque.strand, check_position, AnnMatchCat.INTRON_MID))
                                else:
                                    assert(False), "Not implemented logic"
                            ref_deque.pop_left(safe=True)
                        ref_deque.restore()
                        break

                if len(current_annotations) == 0:                                            # in case we failed to annotate junction
                    logging.debug(f"""Failed to annotate {check_position:,}""")
                    current_annotations.append(None)

                all_annotations.append((current_annotations, int(current_query.score)))      # score keeps the original rows order

    all_annotations.sort(key = lambda i:i[1])    # need to sort the results so the order corresponds to the jun_coords_location
    return all_annotations


def process_jun_annotation(args, job):
    setup_logger(
        multiprocessing.get_logger(),
        args.loglevel
    )
    multiprocessing.current_process().name = job.contig

    logging.info(f"""Annotating junctions coordinates from {args.jun_coords_location}, subset to {job.contig} chromosome""")
    logging.info(f"""Temporary results will be saved to {job.location}""")

    logging.debug("Annotating junctions start coordinates")
    sorted_start_annotations = get_annotation(
        job=job,
        query_location=args.jun_starts_location,
        references_location=args.ref,
        threads=args.threads
    )

    logging.debug("Annotating junctions end coordinates")
    sorted_end_annotations = get_annotation(
        job=job,
        query_location=args.jun_ends_location,
        references_location=args.ref,
        threads=args.threads
    )

    with job.location.open("wt") as output_stream:
        with pysam.TabixFile(str(args.jun_coords_location), mode="r", parser=pysam.asBed(), threads=args.threads) as jun_coords_handler:
            for (current_coords, (start_annotations, _), (end_annotations, _)) in zip(
                jun_coords_handler.fetch(get_correct_contig(job.contig, jun_coords_handler)), sorted_start_annotations, sorted_end_annotations
            ):
                logging.debug(f"""Assigning annotation for {job.contig}:{current_coords.start:,}-{current_coords.end:,}, {current_coords.name}""")
                start_annotations = [_ for _ in start_annotations if _ is not None]
                end_annotations = [_ for _ in end_annotations if _ is not None]
                grouped_annotation = GroupedAnnotations()
                for annotations_pair in list(itertools.product(start_annotations, end_annotations)):   # if any of [start_]/[end_]annotations is [], returns []
                    grouped_annotation.add(annotations_pair[0], annotations_pair[1])
                annotation, strand = grouped_annotation.best()
                output_stream.write(
                    f"""{job.contig}\t{current_coords.start}\t{current_coords.end}\t{current_coords.name}\t{annotation}\t{strand}\n"""   # no reason to keep it BED-formatted
                )


def load_counts_data(query_locations, query_aliases, selected_chr, tmp_location, as_junctions=None, save_bed=None):
    save_bed = False if save_bed is None else save_bed
    loading_params = JunctionsParams if as_junctions is not None and as_junctions else IntronsParams

    counts_df = None
    for query_location, query_alias in zip(query_locations, query_aliases):
        logging.info(f"""Loading counts from {query_location} as {query_alias}""")
        current_df = pandas.read_csv(query_location, **loading_params)
        current_df.rename(
            columns = {"score": query_alias},
            inplace = True
        )
        current_df = current_df[current_df.index.get_level_values("chr").isin(selected_chr)]                         # subset only to those chromosomes that are provided in --chr
        counts_df = current_df if counts_df is None else counts_df.join(current_df, how="outer").fillna(0)
        counts_df = counts_df.astype(pandas.SparseDtype("int32", 0))                                                # saves memory and is required before exporting to h5ad

    logging.info("Sorting counts by coordinates in ascending order")
    counts_df.sort_index(ascending=True, inplace=True)

    counts_location = tmp_location.joinpath(get_tmp_suffix()).with_suffix(".pickle")
    logging.info(f"""Saving counts with coordinates as a temporary pickled file to {counts_location}""")
    counts_df.to_pickle(counts_location)

    coords_location, starts_location, ends_location = None, None, None
    if save_bed:                                                                                                     # no need to allow to save BED when we alreade have intron annotations
        coords_df = counts_df.index.to_frame(index=False)
        coords_df["score"] = range(0, len(coords_df.index))
        coords_location = tmp_location.joinpath(get_tmp_suffix()).with_suffix(".bed")
        logging.info(f"""Saving only coordinates as a temporary BED file to {coords_location}""")
        coords_df.to_csv(
            coords_location,
            sep="\t",
            header=False,
            index=False,
            columns=loading_params["names"]
        )
        coords_location = get_indexed_bed(coords_location, keep_original=False, force=True)

        start_coords_df = coords_df.copy()
        start_coords_df["end"] = start_coords_df["start"]
        start_coords_df.sort_values(
            by=loading_params["index_col"],
            ascending=True,
            inplace=True
        )
        starts_location = tmp_location.joinpath(get_tmp_suffix()).with_suffix(".bed")
        logging.info(f"""Saving only start coordinates as a temporary BED file to {starts_location}""")
        start_coords_df.to_csv(
            starts_location,
            sep="\t",
            header=False,
            index=False,
            columns=loading_params["names"]
        )
        starts_location = get_indexed_bed(starts_location, keep_original=False, force=True)

        end_coords_df = coords_df.copy()
        end_coords_df["start"] = end_coords_df["end"]
        end_coords_df.sort_values(
            by=loading_params["index_col"],
            ascending=True,
            inplace=True
        )
        ends_location = tmp_location.joinpath(get_tmp_suffix()).with_suffix(".bed")
        logging.info(f"""Saving only end coordinates as a temporary BED file to {ends_location}""")
        end_coords_df.to_csv(
            ends_location,
            sep="\t",
            header=False,
            index=False,
            columns=loading_params["names"]
        )
        ends_location = get_indexed_bed(ends_location, keep_original=False, force=True)

    return (counts_location, coords_location, starts_location, ends_location)


def collect_results(args):

    counts_df = None
    if args.juncounts is not None:
        collected_annotations_df = None
        for job in args.jun_annotation_jobs:
            logging.info(f"""Loading annotated junctions coordinates from {job.location}""")
            annotations_df = pandas.read_csv(job.location, **AnnotationsParams)
            collected_annotations_df = annotations_df if collected_annotations_df is None else pandas.concat([collected_annotations_df, annotations_df])
            logging.info(f"""Removing {job.location}""")
            job.location.unlink()
        logging.info(f"""Loading pickled junctions counts from {args.jun_counts_location}""")
        counts_df = pandas.read_pickle(args.jun_counts_location)
        logging.info("Joining pickled junctions counts with annotations")
        counts_df = counts_df.join(collected_annotations_df, how="left")

    int_counts_df = None
    if args.intcounts is not None:
        logging.info(f"""Loading pickled introns counts from {args.int_counts_location}""")
        int_counts_df = pandas.read_pickle(args.int_counts_location)
        int_counts_df["annotation"] = int_counts_df.index.to_frame(index=False).name.values
        int_counts_df.reset_index(level="strand", inplace=True)                                     # need to move "strand" from the index to a column

    if counts_df is None:
        counts_df = int_counts_df
    elif int_counts_df is not None:
        logging.info("Concatenating annotated junctions and introns counts from")
        counts_df = pandas.concat([counts_df, int_counts_df])

    counts_df.sort_index(ascending=True, inplace=True)

    adata_location = args.output.with_suffix(".h5ad")
    logging.info(f"""Exporting aggregated counts to {adata_location}""")
    metadata_columns = ["annotation", "strand"]
    export_counts_to_anndata(
        counts_df=counts_df,
        location=adata_location,
        counts_columns=[c for c in counts_df.columns.values if c not in metadata_columns],
        strand_coords=True,
        metadata_columns=metadata_columns
    )

    if args.bed:
        bed_location = args.output.with_suffix(".bed")
        logging.info(f"""Exporting annotated coordinates to {bed_location}""")
        counts_df.reset_index(inplace=True)
        counts_df.drop(columns=["name"], inplace=True)
        counts_df.rename(columns={"annotation": "name"}, inplace=True)
        counts_df["score"] = 0
        counts_df.to_csv(
            bed_location,
            sep="\t",
            header=False,
            index=False,
            columns=["chr", "start", "end", "name", "score", "strand"]
        )
        bed_location = get_indexed_bed(bed_location, keep_original=False, force=True)


def aggregate(args):

    if args.juncounts is not None:
        logging.info("Processing junctions counts")
        args.jun_counts_location, args.jun_coords_location, args.jun_starts_location, args.jun_ends_location = load_counts_data(
            query_locations=args.juncounts,
            query_aliases=args.aliases,
            selected_chr=args.chr,
            tmp_location=args.tmp,
            as_junctions=True,                                                       # "strand" column will be ignored
            save_bed=True
        )
        args.jun_annotation_jobs = get_jun_annotation_jobs(args)
        logging.info(f"""Span {len(args.jun_annotation_jobs)} junctions annotation job(s) between {args.cpus} CPU(s)""")
        with multiprocessing.Pool(args.cpus) as pool:
            pool.map(partial(process_jun_annotation, args), args.jun_annotation_jobs)

    if args.intcounts is not None:
        logging.info("Processing introns counts")
        args.int_counts_location, _, _, _ = load_counts_data(
            query_locations=args.intcounts,
            query_aliases=args.aliases,
            selected_chr=args.chr,
            tmp_location=args.tmp,
            save_bed=False
        )

    logging.info("Collecting results")
    collect_results(args)

    logging.info("Removing temporary directory")
    shutil.rmtree(args.tmp)


import os
import sys
import csv
import anndata as ad
import pandas as pd
import numpy as np
from scipy.sparse import csr_matrix
from scipy.io import mmread
import collections
sys.path.insert(1, os.path.join(sys.path[0], '..'))
from . import isoform_junctions_chunk as junc
from . import isoform_matrix as iso
from . import isoform_ratios as isor
from . import isoform_automate as isoa
from . import isoform_translation as isot
from . import gff_process as gff_process
from ..psi import psi_single as psi
from Bio import SeqIO
import asyncio

def exportRatios(ob,out):
    pseudo_pdf, tpm, isoform_to_gene_ratio = iso.pseudo_cluster_counts(ob,cell_threshold=5,count_threshold=0,compute_tpm=True)
    pseudo_pdf.to_csv(out+'_pseudo_cluster_counts.txt', sep='\t')
    tpm.to_csv(out+'_pseudo_cluster-tpm.txt', sep='\t')
    isoform_to_gene_ratio.to_csv(out+'_pseudo_cluster-ratio.txt', sep='\t')

def import_metadata(metadata_file):
    """Reads metadata file and groups samples by uid."""
    metadata = pd.read_csv(metadata_file, sep='\t')
    sample_dict = {}
    for _, row in metadata.iterrows():
        uid = row['uid']
        if uid not in sample_dict:
            sample_dict[uid] = []
        sample_dict[uid].append({
            'gff': row['gff'], 
            'gff_name': os.path.basename(row['gff']).split('.')[0],
            'matrix': row['matrix'], 
            'library': row['library'], 
            'reverse': row['reverse'],
            'groups': row['groups']
        })
    return sample_dict

def export_junction_matrix(matrix, gff, library, reverse, ensembl_exon_dir, barcode_sample_dict):
    """Exports junction matrix for a single sample."""
    adata = junc.exportJunctionMatrix(matrix, ensembl_exon_dir, gff, barcode_clusters=barcode_sample_dict[library], rev=reverse)
    return iso.append_sample_name(adata, library)

def export_isoform_matrix(matrix, gff, library, isoform_associations_path, reverse, barcode_sample_dict):
    """Exports an integrated isoform matrix for a single sample."""
    adata = isor.exportConsensusIsoformMatrix(matrix,isoform_associations_path,gff_source=gff,barcode_clusters=barcode_sample_dict[library],rev=reverse)
    return iso.append_sample_name(adata, library)

def export_junction_h5ad(sample_dict, ensembl_exon_dir, barcode_sample_dict):
    """Processes samples, concatenates if necessary, and writes h5ad files."""
    for uid, samples in sample_dict.items():
        num_samples = len(samples)
        print(f'Processing {num_samples} library for sample: {uid}')
        adata_list = [export_junction_matrix(s['matrix'], s['gff'], s['library'], s['reverse'], ensembl_exon_dir, barcode_sample_dict) for s in samples]
        if num_samples>1:
            combined_adata = ad.concat(adata_list, axis=0, join='outer') if len(adata_list) > 1 else adata_list[0]
            combined_adata.write_h5ad(f'{uid}.h5ad', compression='gzip')

def export_isoform_h5ad(sample_dict, ensembl_exon_dir, barcode_sample_dict, reference_gff, genome_fasta):
    """Processes samples, concatenates if necessary, and writes h5ad files."""

    current_dir = os.getcwd()
    gff_output = "/gff-output/"
    query_gff_file = current_dir+gff_output+"combined.gff"
    transcript_associations_path = current_dir+gff_output+"transcript_associations.txt"
    isoform_associations_path = current_dir+gff_output+"isoform_links.txt"
    if os.path.exists(current_dir+gff_output):
        pass
    else:
        # Create a combined GFF
        gff_files = [reference_gff]
        for uid, samples in sample_dict.items():
            for s in samples:
                gff_files.append(s['gff'])
        gff_output_dir = gff_process.consolidateLongReadGFFs(gff_files, ensembl_exon_dir, mode='Ensembl')
        
        # Export isoform translations
        cds_records, transcript_records, protein_records = isot.gff_translate(query_gff_file,genome_fasta,reference_gff,transcript_associations_path)
        with open("protein_sequences.fasta", "w") as protein_file:
            SeqIO.write(protein_records, protein_file, "fasta")
        with open("transcript_sequences.fasta", "w") as cds_file:
            SeqIO.write(transcript_records, cds_file, "fasta")
        with open("cds_sequences.fasta", "w") as cds_file:
            SeqIO.write(cds_records, cds_file, "fasta")

    # Export individual library/sample isoform h5ad
    for uid, samples in sample_dict.items():
        num_samples = len(samples)
        print(f'Processing {num_samples} library for sample: {uid}')
        adata_list = [export_isoform_matrix(s['matrix'], s['gff_name'], s['library'], isoform_associations_path, s['reverse'], barcode_sample_dict) for s in samples]
        if num_samples>1:
            combined_adata = ad.concat(adata_list, axis=0, join='outer') if len(adata_list) > 1 else adata_list[0]
            combined_adata.write_h5ad(f'{uid}_isoform.h5ad')

def get_valid_h5ad(sample_dict,dataType):
    sample_files=[]
    uids=[]
    current_dir = os.getcwd()
    for uid, samples in sample_dict.items():
        h5ad = uid + ('.h5ad' if dataType == 'junction' else '-isoform.h5ad')
        if not os.path.exists(f'{current_dir}/{h5ad}'):
            for s in samples:
                h5ad = s['gff_name'] + ('.h5ad' if dataType == 'junction' else '-isoform.h5ad')
        sample_files.append(h5ad)
        uids.append(uid)
    return sample_files,uids

def get_sample_to_group(sample_dict,dataType):
    sample_group={}
    current_dir = os.getcwd()
    for uid, samples in sample_dict.items():
        sample = uid + ('' if dataType == 'junction' else '-isoform')
        group = samples[0]['groups']
        sample_group[sample]=group
        if not os.path.exists(f'{current_dir}/{sample}'):
            for s in samples:
                sample = s['gff_name'] + ('' if dataType == 'junction' else '-isoform')
                sample_group[sample]=group
    return sample_group

import os

def export_sorted(filename, sort_col, exclude_header=True):
    """
    Efficiently sort a large file without loading everything into memory.
    The file is sorted based on the specified column (0-indexed).
    """
    output_file = filename[:-4] + '-sorted'  # Temporary sorted file
    index = []  # Store tuples of (sort key, offset, line length)

    with open(filename, 'r', encoding='utf-8') as f:
        header = None
        if exclude_header:
            header = f.readline()  # Store header separately
        while True:
            offset = f.tell()  # Track the byte offset of each line
            line = f.readline()
            if not line:
                break  # Stop if end of file is reached
            length = len(line)
            col_value = line.split('\t')[sort_col].strip()  # Extract the sort key
            index.append((col_value, offset, length))  # Add to index
    index.sort()

    with open(output_file, 'w', encoding='utf-8') as o:
        if exclude_header and header:
            o.write(header)  # Write header to the output file
        with open(filename, 'r', encoding='utf-8') as f:
            for col_value, offset, length in index:
                f.seek(offset)  # Move to the line's position
                o.write(f.read(length))  # Write the line to the output
    try:
        os.remove(filename)  # Delete the original file
        os.rename(output_file, filename)  # Rename sorted file to original name
        return filename
    except Exception as e:
        print(f"Error replacing the original file: {e}")
        return output_file

def export_pseudo_counts(metadata_file,barcode_cluster_dirs,dataType='junction',compute_tpm=False):
    sample_dict = isoa.import_metadata(metadata_file)
    sample_files,uids = get_valid_h5ad(sample_dict,dataType)
    cluster_order = iso.return_cluster_order(barcode_cluster_dirs)

    # Memory efficient combine and pseudobulk of many h5ad files
    pseudo_counts = dataType+'_combined_pseudo_cluster_counts.txt'
    pseudo_tpm = dataType+'_combined_pseudo_cluster_tpm.txt'
    pseudo_ratios = dataType+'_combined_pseudo_cluster_ratio.txt'
    
    iso.concatenate_h5ad_and_compute_pseudobulks_optimized(sample_files,collection_name=dataType,compute_tpm=compute_tpm)

    # Organize and filter
    iso.export_and_filter_pseudobulks(pseudo_counts, pseudo_counts[:-4]+'-filtered.txt', cluster_order)
    if dataType == 'isoform':
        iso.export_and_filter_pseudobulks(pseudo_ratios, pseudo_ratios[:-4]+'-filtered.txt', cluster_order)
        iso.export_and_filter_pseudobulks(pseudo_tpm, pseudo_tpm[:-4]+'-filtered.txt', cluster_order)

def pre_process_samples(metadata_file, barcode_cluster_dirs, ensembl_exon_dir):
    # Perform junction quantification across samples
    sample_dict = import_metadata(metadata_file)
    barcode_sample_dict = iso.import_barcode_clusters(barcode_cluster_dirs)
    export_junction_h5ad(sample_dict, ensembl_exon_dir, barcode_sample_dict)
    export_pseudo_counts(metadata_file,barcode_cluster_dirs,'junction')

def combine_processed_samples(metadata_file, barcode_cluster_dirs, ensembl_exon_dir, gencode_gff, genome_fasta):
    sample_dict = import_metadata(metadata_file)
    barcode_sample_dict = iso.import_barcode_clusters(barcode_cluster_dirs)

    # Perform isoform quantification across samples
    export_isoform_h5ad(sample_dict, ensembl_exon_dir, barcode_sample_dict, gencode_gff, genome_fasta)
    export_pseudo_counts(metadata_file,barcode_cluster_dirs,'isoform',compute_tpm=True)
    
    junction_coords_file = 'junction_combined_pseudo_cluster_counts-filtered.txt'
    outdir = 'psi_combined_pseudo_cluster_counts.txt'
    junction_coords_file = export_sorted(junction_coords_file, 0) ### Sort the expression file
    #psi.main(junction_path=junction_coords_file, query_gene=None, outdir=outdir, use_multiprocessing=False, mp_context=mp_context, num_cores=num_cores)
    run_psi_analysis(junction_coords_file, outdir)

# Wrap in a function to avoid event loop conflicts
def run_psi_analysis(junction_path, outdir):
    # Use asyncio.run to ensure event loop is properly created
    asyncio.run(psi.main(junction_path=junction_path, query_gene=None, outdir=outdir))



import os,sys
import pandas as pd
import numpy as np
from copy import deepcopy
from tqdm import tqdm  # For the progress bar
import time
import multiprocessing
import argparse

# Original method for clique finding
def find_uid_in_clique(the_uid, region, strand, uid2coords):
    clique = {the_uid: region}
    region_start, region_end = map(int, region)

    for uid, coords in uid2coords.items():
        if uid != the_uid:
            coords_start, coords_end = map(int, coords)
            # Check overlap
            if strand == '+':
                if max(region_start, coords_start) - min(region_end, coords_end) < 0:
                    clique[uid] = coords
            elif strand == '-':
                if max(region_end, coords_end) - min(region_start, coords_start) < 0:
                    clique[uid] = coords

    return clique


# Early exit conditional checking to optimize
def is_valid(mat, uid):
    num_incl_events = np.count_nonzero(mat[0, :] >= 20)
    if num_incl_events <= 1:
        return False  # Early exit if inclusion events are too low

    num_excl_events = np.count_nonzero(mat[1:, :].sum(axis=0) >= 20)
    if num_excl_events <= 1:
        return False  # Early exit if exclusion events are too low

    total_number_junctions = np.count_nonzero(mat.sum(axis=0) >= 20)
    if total_number_junctions <= 2:
        return False  # Early exit if total junction count is too low

    return True


# Core PSI calculation with early exit conditions
def calculate_psi_core(clique, uid, count, sample_columns):
    sub_count = count.loc[list(clique.keys()), sample_columns]
    mat = sub_count.values
    with np.errstate(divide='ignore', invalid='ignore'):
        psi = mat[0, :] / mat.sum(axis=0)

    if sub_count.shape[0] > 1:
        bg_uid = sub_count.index.tolist()[np.argmax(mat[1:, :].sum(axis=1)) + 1]
        cond = is_valid(mat, uid)
    else:
        bg_uid = 'None'
        cond = False

    return psi, bg_uid, cond


# PSI calculation for a chunk of genes
def calculate_psi_per_gene_chunk(chunk_data):
    count, header, sample_columns, lookup_table = chunk_data
    result_data = []
    uid2coords = count.apply(lambda x: [x['start'], x['end']], axis=1, result_type='reduce').to_dict()

    for row in count.iterrows():
        uid = row[0]  # This should be the correct splice junction ID
        region = uid2coords[uid]
        strand = row[1]['strand']
        clique = find_uid_in_clique(uid, region, strand, uid2coords)
        psi_values, bg_uid, cond = calculate_psi_core(clique, uid, count, sample_columns)

        if np.nanmax(psi_values) - np.nanmin(psi_values) >= 0.1 and cond:
            # Ensure the `uid` splice junction IDs are retained by looking up in the lookup table
            uid_str = lookup_table.get(uid, uid)  # Use lookup table to get original string
            bg_uid_str = lookup_table.get(bg_uid, bg_uid)  # Map bg_uid to string
            data = (uid_str + '|' + bg_uid_str, *psi_values)
            result_data.append(data)

    if result_data:
        df = pd.DataFrame.from_records(result_data, columns=['uid'] + sample_columns)
        return df
    else:
        return None


# Write results to file synchronously (with a single write step after parallel processing)
def write_to_file(outdir, results, write_header=True):
    with open(outdir, 'a') as f:
        for df in results:
            if df is not None:
                df.to_csv(f, sep='\t', index=False, header=write_header)
                write_header = False  # Ensure the header is only written once


# Modify the function signature to accept a multiprocessing context
def process_junctions_in_chunks(junction_path, query_gene, outdir, total_lines, mp_context=None, num_cores=None):
    col_uid = []
    col_gene = []
    col_chrom = []
    col_start = []
    col_end = []
    col_strand = []
    data = []
    current_gene = None
    gene_chunks = []  # Store gene data chunks
    sample_columns = None
    lookup_table = {}  # Dictionary to map index to string IDs

    with tqdm(total=total_lines, desc="Processing junctions") as pbar:
        with open(junction_path, 'r') as f:
            header = next(f).strip().split('\t')
            sample_columns = header[1:]

            for line in f:
                pbar.update(1)
                row = line.strip().split('\t')
                uid, coords = row[0].split('=')  # This should be the correct splice junction ID
                chrom, coords = coords.split(':')
                start, end = coords.split('-')
                strand = '+' if start < end else '-'
                gene = uid.split(':')[0]

                # Store the mapping from index to original string ID in the lookup table
                index = len(lookup_table)  # Assuming index is the current length of the lookup table
                lookup_table[index] = uid  # Map the index to the string ID

                # If processing a specific gene, skip others
                if query_gene and gene != query_gene:
                    continue

                if current_gene and current_gene != gene:
                    count = pd.DataFrame(data, columns=header[1:])
                    for name, col in zip(['uid', 'gene', 'chrom', 'start', 'end', 'strand'],
                                         [col_uid, col_gene, col_chrom, col_start, col_end, col_strand]):
                        count[name] = col

                    # Collect chunks for multiprocessing
                    gene_chunks.append((count, header, sample_columns, lookup_table))

                    # Reset collections for the next gene
                    col_uid, col_gene, col_chrom, col_start, col_end, col_strand, data = [], [], [], [], [], [], []

                current_gene = gene
                col_uid.append(uid)  # Properly track splice junction IDs
                col_gene.append(gene)
                col_chrom.append(chrom)
                col_start.append(start)
                col_end.append(end)
                col_strand.append(strand)
                data.append([float(x) if x.replace('.', '', 1).isdigit() else np.nan for x in row[1:]])

            if data:
                count = pd.DataFrame(data, columns=header[1:])
                for name, col in zip(['uid', 'gene', 'chrom', 'start', 'end', 'strand'],
                                     [col_uid, col_gene, col_chrom, col_start, col_end, col_strand]):
                    count[name] = col
                gene_chunks.append((count, header, sample_columns, lookup_table))

    # Use the provided multiprocessing context to process gene chunks
    if mp_context and num_cores:
        with mp_context.Pool(num_cores) as pool:
            results = pool.map(calculate_psi_per_gene_chunk, gene_chunks)
    else:
        results = [calculate_psi_per_gene_chunk(chunk) for chunk in gene_chunks]

    # Write results in a single pass
    write_to_file(outdir, results, write_header=True)


def main(junction_path=None, query_gene=None, outdir=None, use_multiprocessing=True, mp_context=None, num_cores=None):
    print(f"Computing PSI from: {junction_path}")
    if junction_path is None or outdir is None:
        raise ValueError("Required parameters: junction_path and outdir must be provided.")

    # Count total lines for progress bar
    total_lines = sum(1 for line in open(junction_path))

    # Initialize the output file (overwrite if it exists)
    with open(outdir, 'w') as f:
        pass  # Simply open and close the file to overwrite/clear it

    # Process junctions in chunks with or without multiprocessing
    start_time = time.time()

    process_junctions_in_chunks(junction_path, query_gene, outdir, total_lines, mp_context, num_cores)

    end_time = time.time()
    print(f"Total time: {end_time - start_time:.2f} seconds")


# Protect multiprocessing entry point
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Compute PSI from count junction')
    parser.add_argument('--junction', type=str, default=None, help='the path to the junction count')
    parser.add_argument('--gene', type=str, default=None, help='gene you want to compute PSI')
    parser.add_argument('--outdir', type=str, default=None, help='output dir for the output file')
    args = parser.parse_args()

    # Create a multiprocessing context
    from multiprocessing import get_context, cpu_count
    mp_context = get_context('spawn')

    # Call main with parsed arguments and multiprocessing enabled
    main(junction_path=args.junction, query_gene=args.gene, outdir=args.outdir, use_multiprocessing=True, mp_context=mp_context, num_cores=cpu_count())



import math

import scanpy as sc
import numpy as np
from sklearn.svm import OneClassSVM
import scanpy.external as sce
from sklearn.ensemble import IsolationForest

from sklearn import metrics
import matplotlib.pyplot as plt
from sklearn.metrics.pairwise import cosine_similarity


def gene_cosine_similarity(train_adata, test_adata):

    # for plotting purposes
    n_cells_train = np.shape(train_adata)[0]
    sc.tl.pca(train_adata, n_comps=min(30, n_cells_train - 1))
    sc.tl.pca(test_adata, n_comps=min(30, n_cells_train - 1))

    # compute similarity matrix
    similarity_matrix = cosine_similarity(test_adata.X, train_adata.X)
    print("computed similarity matrix", flush=True)

    # average "similarity score" for each cell to all the controls
    row_mean_similarity = np.mean(similarity_matrix, axis=1)

    predicted_labels = np.zeros(np.shape(similarity_matrix)[0])
    predicted_labels[row_mean_similarity < np.mean(row_mean_similarity)] = 1

    return train_adata, test_adata, predicted_labels, row_mean_similarity


def one_class_svm(train_adata, test_adata, nu=0.1, matrix="gex"):

    model = OneClassSVM(gamma='scale', nu=nu)

    if matrix == "gex":

        model.fit(train_adata.X)
        y_pred = model.predict(test_adata.X)
        decision_values = model.decision_function(test_adata.X)

    elif matrix == "pca":
        n_cells_train = np.shape(train_adata)[0]
        sc.tl.pca(train_adata, n_comps=min(30, n_cells_train - 1))
        sc.tl.pca(test_adata, n_comps=min(30, n_cells_train - 1))

        pca_train = train_adata.obsm['X_pca']
        pca_test = test_adata.obsm['X_pca']

        model.fit(pca_train)
        y_pred = model.predict(pca_test)
        decision_values = model.decision_function(pca_test)

    elif matrix == "harmony_pca":
        n_cells_train = np.shape(train_adata)[0]
        sc.tl.pca(train_adata, n_comps=min(30, n_cells_train - 1))
        nclust_harmony = np.min([np.round(n_cells_train/ 30.0), 100]).astype(int)
        if nclust_harmony <= 1:
            nclust = 2
        else:
            nclust = None
        sce.pp.harmony_integrate(train_adata, 'patient_id', nclust=nclust)
        harmony_pca_train = train_adata.obsm['X_pca_harmony']

        sc.tl.pca(test_adata, n_comps=np.shape(harmony_pca_train)[1])
        pca_test = test_adata.obsm['X_pca']
        # harmony_pca_test = test_adata.obsm['X_pca_harmony']

        model.fit(harmony_pca_train)
        y_pred = model.predict(pca_test)
        decision_values = model.decision_function(pca_test)

    # convert the outliers (-1) to disease (1)
    y_pred[y_pred == 1] = 0
    y_pred[y_pred == -1] = 1
    predicted_labels = y_pred

    return train_adata, test_adata, predicted_labels, decision_values


def isolation_forest(train_adata, test_adata, nu=0.1, matrix="gex"):

    model = IsolationForest()

    if matrix == "gex":

        model.fit(train_adata.X)
        y_pred = model.predict(test_adata.X)
        decision_values = model.decision_function(test_adata.X)

    elif matrix == "pca":
        n_cells_train = np.shape(train_adata)[0]
        sc.tl.pca(train_adata, n_comps=min(30, n_cells_train - 1))
        sc.tl.pca(test_adata, n_comps=min(30, n_cells_train - 1))

        pca_train = train_adata.obsm['X_pca']
        pca_test = test_adata.obsm['X_pca']

        model.fit(pca_train)
        y_pred = model.predict(pca_test)
        decision_values = model.decision_function(pca_test)

    elif matrix == "harmony_pca":
        n_cells_train = np.shape(train_adata)[0]
        sc.tl.pca(train_adata, n_comps=min(30, n_cells_train - 1))
        nclust_harmony = np.min([np.round(n_cells_train/ 30.0), 100]).astype(int)
        if nclust_harmony <= 1:
            nclust = 2
        else:
            nclust = None
        sce.pp.harmony_integrate(train_adata, 'patient_id', nclust=nclust)
        harmony_pca_train = train_adata.obsm['X_pca_harmony']

        sc.tl.pca(test_adata, n_comps=np.shape(harmony_pca_train)[1])
        pca_test = test_adata.obsm['X_pca']
        # harmony_pca_test = test_adata.obsm['X_pca_harmony']

        model.fit(harmony_pca_train)
        y_pred = model.predict(pca_test)

    # convert the outliers (-1) to disease (1)
    y_pred[y_pred == 1] = 0
    y_pred[y_pred == -1] = 1
    predicted_labels = y_pred

    return predicted_labels, decision_values


def clf_metrics_all(true_labels, predicted_labels):

    # Compute the ROC curve
    fpr, tpr, _ = metrics.roc_curve(true_labels, predicted_labels)

    # Compute the AUC (Area Under the Curve) value
    auc = metrics.roc_auc_score(true_labels, predicted_labels)

    # Compute the confusion matrix
    conf_matrix = metrics.confusion_matrix(true_labels, predicted_labels)
    specificity_healthy = conf_matrix[0, 0]/(conf_matrix[0, 0] + conf_matrix[0, 1])
    sensitivity_disease = conf_matrix[1, 1] / (conf_matrix[1, 0] + conf_matrix[1, 1])

    # Generate the classification report
    classification_report = metrics.classification_report(true_labels, predicted_labels, target_names=['Healthy', 'Disease'])

    return fpr, tpr, auc, specificity_healthy, sensitivity_disease, conf_matrix, classification_report


def find_top_bottom_cells(metadata, decision_obs_col, top_perc=0.15, bottom_perc=0.15):

    metadata = metadata.sort_values(by=[decision_obs_col])
    metadata["new_predicted_labels"] = 2
    n_cells = np.shape(metadata)[0]
    top_n_cells = math.ceil(top_perc*n_cells)
    bottom_n_cells = math.ceil(bottom_perc*n_cells)

    metadata.iloc[:top_n_cells, 13] = 1   # disease
    metadata.iloc[-bottom_n_cells:, 13] = 0  # healthy

    return metadata



import pandas as pd
import numpy as np
from scipy.stats import fisher_exact
import math
from statsmodels.stats.multitest import multipletests


def call_pathway_database(species, database_file_path=None):

    # read gtf file
    if database_file_path is not None:
        # read gtf file
        pathways_df = read_database_file(database_file_path)
    else:
        # Filter based on species
        if species == "Hs":
            pathways_df = pd.read_csv('PathwayCommons-Hs.txt', sep="\t", names=["GeneID", "Pathway"], header=0)
            # Filter uppercase entries or those starting with "ENSG"
            pathways_df = pathways_df[(pathways_df['GeneID'].str.isupper()) | (pathways_df['GeneID'].str.startswith('ENSG'))]
        elif species == "Mm":
            # Filter uppercase entries or those starting with "ENSG"
            pathways_df = pd.read_csv('PathwayCommons-Mm.txt', sep="\t", names=["GeneID", "Pathway"], header=0)
            pathways_df = pathways_df[~(pathways_df['GeneID'].str.isupper()) | (pathways_df['GeneID'].str.startswith('ENSMU'))]

    return pathways_df


def read_database_file(pathways_database_file_path):

    columns = ["GeneID", "Pathway"]
    df = pd.read_csv(pathways_database_file_path, sep='\t', comment='#', names=columns, dtype={"seqname": str}, header=0)

    return df


# Define functions for Z-score and Fisher's exact test
def calculate_z_score_aa(associated_in_group, in_static_interval, total, in_flexible_interval):
    r = float(associated_in_group)  # genes regulated in pathway
    _n = float(in_static_interval)  # genes regulated
    N = float(total)  # measured genes in the genome - total_count (#all genes evaluated on pathways)
    R = float(in_flexible_interval)  #genes in pathway
    if (R - N) == 0:
        return 0
    elif r == 0 and _n == 0:
        return 0
    else:
        try:
            # try:
            z = (r - _n * (R / N)) / math.sqrt(_n * (R / N) * (1 - (R / N)) * (1 - ((_n - 1) / (N - 1))))
            return z
            # except ZeroDivisionError: print 'r,_n,R,N: ', r,_n,R,N;kill
        except ValueError:
            print(r - _n * (R / N)), _n * (R / N) * (1 - (R / N)) * (1 - ((_n - 1) / (N - 1))), r, _n, N, R;kill
    # if expected == 0:
        # return 0
    # return (observed - expected) / np.sqrt(expected * (1 - (expected / total_genes)))


def calculate_z_score(observed, expected, total_genes):
    if expected == 0:
        return 0
    return (observed - expected) / np.sqrt(expected * (1 - (expected / total_genes)))


def fisher_test(contingency_table):

    _, p_value = fisher_exact(contingency_table, alternative='greater')
    return p_value


def pathway_enrichment(adata, pathways_df):

    # Load gene sets DataFrame
    gene_sets_df = adata.uns['udon_marker_genes_top_n']
    gene_sets_df = gene_sets_df[['marker', 'top_cluster']]

    # Convert DataFrame to dictionary
    query_gene_sets = gene_sets_df.groupby('top_cluster')['marker'].apply(list).to_dict()

    # Calculate pathway statistics for each query set
    results = []

    total_genes = pathways_df['GeneID'].nunique()
    total_pathways = pathways_df['Pathway'].nunique()

    for set_name, gene_list in query_gene_sets.items():
        query_gene_df = pathways_df[pathways_df['GeneID'].isin(gene_list)]
        pathway_counts = query_gene_df['Pathway'].value_counts()

        for pathway in pathways_df['Pathway'].unique():
            observed = pathway_counts.get(pathway, 0)
            total_in_pathway = pathways_df[pathways_df['Pathway'] == pathway]['GeneID'].nunique()
            expected = (total_in_pathway / total_genes) * len(gene_list)

            z_score = calculate_z_score(observed, expected, total_genes)

            # Contingency table for Fisher's exact test
            not_in_pathway = total_genes - total_in_pathway
            in_query_set_not_in_pathway = len(gene_list) - observed
            contingency_table = [[observed, total_in_pathway - observed],
                                 [in_query_set_not_in_pathway, not_in_pathway - in_query_set_not_in_pathway]]

            p_value = fisher_test(contingency_table)
            results.append({
                'query_set': set_name,
                'pathway': pathway,
                'observed': observed,
                'expected': expected,
                'z_score': z_score,
                'p_value': p_value
            })


    # Save the results
    results_df = pd.DataFrame(results)

    return results_df


def enrichment_wrapper(adata, species, database_file_path=None, p_val=0.05):

    # Load the database
    pathways_df = call_pathway_database(species, database_file_path)

    # Perform pathway enrichment analysis
    results_df = pathway_enrichment(adata, pathways_df)

    # Apply Benjamini-Hochberg FDR correction within each query gene set
    def apply_fdr_correction(group):
        p_values = group['p_value'].values
        corrected_p_values = multipletests(p_values, method='fdr_bh')[1]
        group['adj_p_value'] = corrected_p_values
        return group

    # Group by 'query_gene_set' and apply the correction
    results_df = results_df.groupby('query_set').apply(apply_fdr_correction).reset_index(drop=True)
    results_df_clean = results_df[results_df['p_value'] < p_val]
    results_df_clean = results_df_clean[results_df_clean['observed'] > 2]

    # Function to get top 5 entries with p_val < 0.05
    def top_n_entries(group, n=5):
        # Filter entries with p_val < 0.05 and get top n by p_val
        return group.nlargest(n, 'z_score')

    # Apply the function to each group
    top_entries = results_df_clean.groupby('query_set').apply(top_n_entries).reset_index(drop=True)

    top_entries = top_entries[top_entries['observed'] > 2]

    adata.uns['significant_pathway_enrichment'] = results_df
    adata.uns['significant_pathway_enrichment_top_n'] = top_entries

    return adata



import re, sys, os
import warnings
from Bio import BiopythonWarning
warnings.simplefilter('ignore', BiopythonWarning)
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio import SeqIO
from tqdm import tqdm
import pandas as pd
from collections import defaultdict

def normalize_chromosome_name(chromosome):
    return chromosome.replace('chr', '')

def parse_gff(gff_file):
    transcripts = {}
    with open(gff_file) as file:
        lines = file.readlines()
        for line in tqdm(lines, desc="Parsing GFF"):
            if line.startswith("#"):
                continue
            parts = line.strip().split('\t')
            if len(parts) < 9:
                continue
            feature_type = parts[2]
            attributes = parts[8]
            if feature_type == "exon":
                match = re.search(r'transcript_id[ =]"?([^";]+)"?', attributes)
                gene_match = re.search(r'gene_id[ =]"?([^";]+)"?', attributes)
                if match:
                    transcript_id = match.group(1)
                    gene_id = gene_match.group(1) if gene_match else ""
                    if transcript_id not in transcripts:
                        transcripts[transcript_id] = {"exons": [], "gene_id": gene_id}
                    normalized_chr = normalize_chromosome_name(parts[0])
                    transcripts[transcript_id]["exons"].append((normalized_chr, int(parts[3]), int(parts[4]), parts[6]))
    print(f"{len(transcripts)} transcripts imported from the GFF")
    return transcripts

def get_longest_orf(seq):
    max_orf = ''
    max_orf_start = -1
    strand, nuc = (+1, seq)  # Only consider the 5' to 3' sequence
    for frame in range(3):
        trans = str(nuc[frame:].translate(to_stop=False))
        trans_len = len(trans)
        aa_start = 0
        while aa_start < trans_len:
            aa_start = trans.find('M', aa_start)  # Find the next start codon
            if aa_start == -1:
                break
            aa_end = trans.find('*', aa_start)
            if aa_end == -1:
                aa_end = trans_len
            orf = trans[aa_start:aa_end]
            if len(orf) > len(max_orf):
                max_orf = orf
                max_orf_start = aa_start * 3 + frame
            aa_start = aa_end + 1
    return max_orf, max_orf_start

def is_nmd(transcript_seq, stop_codon_pos, last_exon_start, start_codon_pos, penultimate_exon_start):
    if stop_codon_pos < last_exon_start - 50:
        #"start codon condition", decreased accuracy when tested
        """if stop_codon_pos < start_codon_pos + 100:
           #return False
        else:
        """
        return True
    #predicts NMD when PTC is more than 2 exons away from end of transcript
    elif stop_codon_pos < penultimate_exon_start:
        return True
    return False

def extract_cds_and_protein(transcripts, genome_fasta, ref_first_exons=None, query_transcript_to_gene={}):
    genome_seq = SeqIO.to_dict(SeqIO.parse(genome_fasta, "fasta"))
    cds_records = []
    transcript_records = []
    protein_records = []
    for transcript_id, data in tqdm(transcripts.items(), desc="Processing Transcripts"):
        #if transcript_id=='ENST00000356073.8':
        exons = data["exons"]
        gene_id = data["gene_id"]
        if transcript_id in query_transcript_to_gene:
            gene_id = query_transcript_to_gene[transcript_id]
        exons.sort(key=lambda x: x[1])
        try:
            if exons[0][3] == '+':
                transcript_seq_list = [str(genome_seq[exon[0]].seq[exon[1]-1:exon[2]]) for exon in exons]
            else:
                transcript_seq_list = [str(genome_seq[exon[0]].seq[exon[1]-1:exon[2]].reverse_complement()) for exon in exons]
                transcript_seq_list = transcript_seq_list[::-1]  # Reverse for negative strand
            transcript_seq = ''.join(transcript_seq_list)

            ref_cds_seq = None
            ref_cds_start_pos = None
            if gene_id in ref_first_exons:
                ref_cds_seq = []
                for (ref_cds_start, ref_cds_end) in ref_first_exons[gene_id]:
                    ref_cds_seq_instance = str(genome_seq[exons[0][0]].seq[ref_cds_start-1:ref_cds_end])
                    if exons[0][3] == '-':
                        ref_seq = Seq(ref_cds_seq_instance)
                        ref_cds_seq_instance = str(ref_seq.reverse_complement())
                    ref_cds_seq.append(ref_cds_seq_instance)

            # Restrict to ORFs with defined start codons if transcript has matching starts
            orf_seq = None
            if ref_cds_seq:
                candidates=[]
                for ref_cds_seq_instance in ref_cds_seq:
                    if len(ref_cds_seq_instance) >= 10:
                        if ref_cds_seq_instance in transcript_seq:
                            orf_start = transcript_seq.find(ref_cds_seq_instance)
                            transcript_seq_alt = Seq(transcript_seq[orf_start:])
                            orf_seq = str(transcript_seq_alt.translate(to_stop=True))
                            candidates.append((len(orf_seq),orf_start,orf_seq))
                if len(candidates)>0:
                    candidates.sort()
                    candidates.reverse()
                    ol,orf_start,orf_seq = candidates[0] # select the most 5' start site

            if orf_seq == None:
                orf_seq, orf_start = get_longest_orf(Seq(transcript_seq))

            if orf_seq:
                cds_seq = transcript_seq[orf_start:orf_start + len(orf_seq) * 3]
                description = f";CDS;gene_id:{gene_id}" if gene_id else "CDS"
                cds_records.append(SeqRecord(Seq(cds_seq), id=transcript_id, description=description))
                description = f";transcript;gene_id:{gene_id}" if gene_id else "CDS"
                transcript_records.append(SeqRecord(Seq(transcript_seq), id=transcript_id, description=description))
                stop_codon_pos = orf_start + len(orf_seq) * 3
                nmd_annotation = ""
                if exons[0][3] == '-':
                    last_exon_length = exons[0][2] - exons[0][1] + 1
                    try: penultimate_exon_length = exons[1][2] - exons[1][1] + 11
                    except: penultimate_exon_legth = 0
                else:
                    last_exon_length = exons[-1][2] - exons[-1][1] + 1
                    try: penultimate_exon_length = exons[-2][2] - exons[-2][1] + 1
                    except: penultimate_exon_length = 0
                last_exon_start_relative = len(transcript_seq) - last_exon_length
                penultimate_exon_start_relative = last_exon_start_relative - penultimate_exon_length #check if last intron length needs to be included?
                nmd_annotation = ";(NMD)" if is_nmd(transcript_seq, stop_codon_pos, last_exon_start_relative, orf_start, penultimate_exon_start_relative) else ""
                protein_description = f";Protein';'gene_id:{gene_id}{nmd_annotation}" if gene_id else f"Protein{nmd_annotation}"
                protein_records.append(SeqRecord(Seq(orf_seq), id=transcript_id, description=protein_description))
                #print (orf_start,(len(orf_seq) * 3),last_exon_length,last_exon_start_relative,stop_codon_pos,transcript_seq,[nmd_annotation])
        except KeyError as e:
            print(f"Error: {e}. Check if chromosome identifiers match between GFF and genome FASTA.")
    return cds_records, transcript_records, protein_records

def parse_transcript_associations(file_path):
    df = pd.read_csv(file_path, sep='\t', header=None)
    df.columns = ['gene_id', 'strand', 'exon_structure', 'transcript_id', 'source']
    transcript_to_gene = {}
    intron_retention_dict = {}

    for _, row in tqdm(df.iterrows(), total=df.shape[0], desc="Parsing Transcript Associations"):
        new_transcript_id = f"{row['source']}_{row['transcript_id']}"
        transcript_to_gene[new_transcript_id] = row['gene_id']
        transcript_to_gene[row['transcript_id']] = row['gene_id']

        exon_structure = row['exon_structure'].split('|')
        intron_retention = any(part.startswith('I') and '_' not in part for part in exon_structure)
        intron_retention_dict[new_transcript_id] = intron_retention
        intron_retention_dict[row['transcript_id']] = intron_retention

    return transcript_to_gene, intron_retention_dict

def get_reference_first_exons(ref_gff_file):
    ref_first_exons = {}
    transcript_cds = {}
    with open(ref_gff_file) as file:
        for line in tqdm(file, desc="Parsing Reference GFF"):
            if line.startswith("#"):
                continue
            parts = line.strip().split('\t')
            if len(parts) < 9 or parts[2] != "CDS":
                continue
            attributes = parts[8]
            gene_match = re.search(r'gene_id[ =]"?([^";]+)"?', attributes)
            transcript_match = re.search(r'transcript_id[ =]"?([^";]+)"?', attributes)
            if gene_match:
                gene_id = gene_match.group(1).split('.')[0]
                transcript_id = transcript_match.group(1).split('.')[0]
                cds_start = int(parts[3])
                cds_end = int(parts[4])
                if transcript_id not in transcript_cds: # Ensure it is the first exon for that transcript
                    try: 
                        if (cds_start, cds_end) not in ref_first_exons[gene_id]:
                            ref_first_exons[gene_id].append((cds_start, cds_end))
                    except: ref_first_exons[gene_id] = [(cds_start, cds_end)]
                transcript_cds[transcript_id]=[]

    print(f"{len(ref_first_exons)} genes imported from the reference GFF")
    return ref_first_exons

def gff_translate(query_gff_file, genome_fasta, ref_gff_file=None, transcript_associations_file=None):
    query_transcripts = parse_gff(query_gff_file)
    
    query_transcript_to_gene, intron_retention_dict = parse_transcript_associations(transcript_associations_file) if os.path.exists(transcript_associations_file) else {}
    ref_first_exons = get_reference_first_exons(ref_gff_file) if os.path.exists(ref_gff_file) else {}
    cds_records, transcript_records, protein_records = extract_cds_and_protein(query_transcripts, genome_fasta, query_transcript_to_gene=query_transcript_to_gene, ref_first_exons=ref_first_exons)
    
    # Output protein information to CSV
    export_protein_summary(protein_records, intron_retention_dict, "protein_summary.txt")

    return cds_records, transcript_records, protein_records

def export_protein_summary(protein_records, intron_retention_dict, output_csv_file):
    summary_data = []
    gene_to_longest_isoform = {}

    # First pass to determine the longest isoform for each gene
    for record in protein_records:
        transcript_id = record.id
        description = record.description
        seq_len = len(record.seq)
        gene_id_match = re.search(r'gene_id:([^ ]+)', description)
        gene_id = gene_id_match.group(1) if gene_id_match else ""

        if gene_id:
            if gene_id not in gene_to_longest_isoform or seq_len > gene_to_longest_isoform[gene_id]['length']:
                gene_to_longest_isoform[gene_id] = {'transcript_id': transcript_id, 'length': seq_len}

    # Second pass to create the summary data
    for record in protein_records:
        transcript_id = record.id
        description = record.description
        prot_seq = record.seq
        aa_start = prot_seq.find('M', 0)
        seq_len = len(prot_seq[aa_start:])
        gene_id_match = re.search(r'gene_id:([^ ]+)', description)
        gene_id = gene_id_match.group(1) if gene_id_match else ""
        nmd_status = "Potential-NMD" if "NMD" in description else "Not-NMD"
        if seq_len<10:
            nmd_status = "NMD"
        intron_retention = intron_retention_dict.get(transcript_id, False)

        longest_isoform_length = gene_to_longest_isoform[gene_id]['length'] if gene_id in gene_to_longest_isoform else "NA"

        summary_data.append([
            gene_id, 
            transcript_id,
            seq_len, 
            nmd_status, 
            intron_retention, 
            longest_isoform_length])

    summary_df = pd.DataFrame(summary_data, columns=[
       "Gene ID",  "Transcript ID", "Protein Length", "NMD Status", "Intron Retention", "Longest Isoform Length"])
    summary_df.to_csv(output_csv_file, index=False, sep='\t')

def find_redundant_proteins(fasta_file, output_file, restrict=None):
    # Dictionary to hold sequences as keys and list of corresponding protein IDs as values
    sequence_dict = defaultdict(list)
    
    # Read the FASTA file and populate the dictionary
    for record in SeqIO.parse(fasta_file, "fasta"):
        sequence = str(record.seq)
        protein_id = record.id
        if restrict is not None:
            if restrict in protein_id:
                sequence_dict[sequence].append(protein_id)
        else:
            sequence_dict[sequence].append(protein_id)
    
    # Write the output to a two-column text file
    with open(output_file, 'w') as out_f:
        for sequence, protein_ids in sequence_dict.items():
            isoform_ids=[]
            for isoform in protein_ids:
                isoform = isoform.split(';')[0]
                if 'PB.' in isoform:
                    isoform = 'PB'+isoform.split('_PB')[1]
                isoform_ids.append(isoform)
            gene_id = protein_ids[0].split(';')[2]  # Assuming gene_id is the prefix before '|' in protein_id
            gene_id = gene_id.split(':')[1]
            redundant_ids = ','.join(isoform_ids)
            out_f.write(f"{gene_id}\t{redundant_ids}\t{sequence}\n")

    print(f"Output written to {output_file}")

if __name__ == '__main__':

    fasta_file = "protein_sequences.fasta"
    #fasta_file = 'cds_sequences.fasta'
    #fasta_file = 'transcript_sequences.fasta'
    output_file = "redundant_proteins.txt"
    #output_file = "redundant_cDNA.txt"
    #output_file = "redundant_transcripts.txt"
    restrict = 'ENSG00000109689'
    find_redundant_proteins(fasta_file, output_file, restrict=restrict);sys.exit()

    # Example usage
    ref_gff_file = "/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/MANE.GRCh38.v0.5.select_ensembl_genomic.gff"
    query_gff_file = "/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/gff-output/combined.gff"
    genome_fasta = "/Users/saljh8/Dropbox/Revio/Other/Variants/SNV/genome.fa"
    transcript_associations_file = "/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/gff-output/transcript_associations.txt"

    cds_records, protein_records = gff_translate(
        query_gff_file, 
        genome_fasta, 
        ref_gff_file=ref_gff_file, 
        transcript_associations_file=transcript_associations_file
    )

    # Output results
    with open("cds_sequences.fasta", "w") as cds_file:
        SeqIO.write(cds_records, cds_file, "fasta")

    with open("protein_sequences.fasta", "w") as protein_file:
        SeqIO.write(protein_records, protein_file, "fasta")


import os
import sys
import csv
import anndata as ad
import pandas as pd
import collections
sys.path.insert(1, os.path.join(sys.path[0], '..'))
import long_read.isoform_matrix as iso
import long_read.isoform_automate as isoa
import long_read.comparisons as comp

#import multiprocessing
# Create multiprocessing context using 'spawn'
#mp_context = multiprocessing.get_context('spawn')
#num_cores = mp_context.cpu_count()

if __name__ == '__main__':
    # Load local database files
    ensembl_exon_dir = '/Users/saljh8/Documents/GitHub/altanalyze/AltDatabase/EnsMart91/ensembl/Hs/Hs_Ensembl_exon.txt'
    gene_symbol_file = '/Users/saljh8/Documents/GitHub/altanalyze/AltDatabase/EnsMart91/ensembl/Hs/Hs_Ensembl-annotations.txt'
    gencode_gff = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/gencode.v45.annotation.gff3'
    genome_fasta = "/Users/saljh8/Dropbox/Revio/Other/Variants/SNV/genome.fa"

    # Load cell barcode to cluster associations (order in terms of similar clusters - if possible)
    bc_dir1 = '/Users/saljh8/Dropbox/Revio/Young-Healthy-Revio/Illumina/SoupX-hg38/cellHarmony-labels-young-old.txt'
    bc_dir2 = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/QueryGroups.cellHarmony.txt'
    bc_dir3 = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND251_MAS-Seq/Illumina/SoupX-hg38/cellHarmony/QueryGroups.cellHarmony.txt'
    bc_dir4 = '/Users/saljh8/Dropbox/Revio/Aged-Healthy-Revio/Illumina/SoupX-hg38/cellHarmony/QueryGroups.cellHarmony.txt'
    bc_dir5 = '/Users/saljh8/Dropbox/Revio/MDS-Revio/Aggregate/QueryGroups.cellHarmony.txt'
    barcode_cluster_dirs = [bc_dir1,bc_dir2,bc_dir3,bc_dir4,bc_dir5]
    # Union of all imported clusters (in identified order)
    cluster_order = iso.return_cluster_order(barcode_cluster_dirs)

    # Load sample metadata
    metadata_file = '/Users/saljh8/Documents/GitHub/altanalyze3/altanalyze3/components/tests/BM_metadata.txt'
    sample_dict = isoa.import_metadata(metadata_file)

    # Pre-process all samples in the metadata file
    isoa.pre_process_samples(metadata_file, barcode_cluster_dirs, ensembl_exon_dir)

    # Integrate all sample processed results into combined junction/splicing, isoform, and isoform ratio files
    isoa.combine_processed_samples(metadata_file, barcode_cluster_dirs, ensembl_exon_dir, gencode_gff, genome_fasta)

    # Perform pairwise comparisons between groups and cell-types
    condition1 = 'young'
    condition2 = 'aged'
    condition3 = 'MDS-pre'
    conditions = [(condition1,condition2),(condition2,condition3)]
    #cluster_order = ['HSC-1','MEP'] ### Only consider specific cell type rather than all
    analyses = ['junction','isoform','isoform-ratio']
    comp.compute_differentials(sample_dict,conditions,cluster_order,gene_symbol_file,analyses=analyses)



import os,sys
import pandas as pd
import glob
from tqdm import tqdm

# File paths
transcript_assoc_file = 'gff-output/transcript_associations.txt'
gene_symbol_file = '/Users/saljh8/Documents/GitHub/altanalyze/AltDatabase/EnsMart91/ensembl/Hs/Hs_Ensembl-annotations.txt'
junction_coords_file = 'junction_combined_pseudo_cluster_counts-filtered.txt'
protein_summary_file = 'protein_summary.txt'
stats_folder = 'dPSI-events'

# 1. Load transcript associations into a nested dictionary {gene: {junction: [isoforms]}}
def load_transcript_associations(file_path, dataType='junction'):
    transcript_dict = {}
    with open(file_path, 'r') as f:
        for line in f:
            gene, _, exon_sequence, isoform, _ = line.strip().split('\t')
            exons = exon_sequence.split('|')
            # Generate junctions by pairing adjacent exons
            junctions = [f"{exons[i]}-{exons[i+1]}" for i in range(len(exons) - 1)]
            # Create an entry for the gene if it doesn't exist
            if gene not in transcript_dict:
                transcript_dict[gene] = {}

            # Associate the isoform with each junction
            if dataType == 'junction':
                for junction in junctions:
                    if junction not in transcript_dict[gene]:
                        transcript_dict[gene][junction] = []
                    transcript_dict[gene][junction].append(isoform)
            else:
                transcript_dict[gene][isoform] = exon_sequence
    return transcript_dict

# 2. Load gene symbols {Ensembl ID: Gene Symbol}
def load_gene_symbols(file_path):
    gene_symbol_dict = {}
    with open(file_path, 'r') as f:
        for line in f:
            parts = line.strip().split('\t')
            if len(parts) >= 2:  # Ensure there are at least two columns
                ensembl_id, gene_symbol = parts[:2]
                gene_symbol_dict[ensembl_id] = gene_symbol
    return gene_symbol_dict


# 3. Load junction coordinates {junction_id: coordinates}
def load_junction_coordinates(file_path):
    coord_dict = {}
    with open(file_path, 'r') as f:
        for line in f:
            feature, _ = line.strip().split('\t', 1)
            try: junction_id, coords = feature.split('=')
            except:
                continue # header
            coord_dict[junction_id] = coords
    return coord_dict

# 4. Load protein summary into a dictionary {isoform: (length, NMD status)}
def load_protein_summary(file_path):
    protein_dict = {}
    with open(file_path, 'r') as f:
        for line in f:
            gene_id, isoform, length, nmd_status, _, longest_length = line.strip().split('\t')
            if '_' in isoform:
                isoform = isoform.split('_')[1]  # Remove dataset prefix
            try: length = int(length)
            except:
                continue
            protein_dict[gene_id,isoform] = (length, nmd_status, longest_length)
    return protein_dict

# 5. Select the longest isoform for each junction
def get_longest_isoform(gene,isoforms, protein_dict):
    longest_isoform = None
    max_length = -1
    for isoform in isoforms:
        length, _, _ = protein_dict.get((gene,isoform), (0, None, None))
        #print(isoform,[length]);sys.exit()
        if length > max_length:
            max_length = length
            longest_isoform = f"{isoform}|{length}"
    return longest_isoform or 'N/A'

# 6. Annotate a stats file
def annotate_junction_stats_file(file_path, transcript_dict, gene_symbol_dict, coord_dict, protein_dict, output_folder):
    stats_df = pd.read_csv(file_path, sep='\t')

    annotations = []  # Store annotated rows

    for index, row in stats_df.iterrows():
        feature = row['Feature']
        junction1, junction2 = feature.split('|')

        gene_id = junction1.split(':')[0]  # Extract gene ID
        gene_symbol = gene_symbol_dict.get(gene_id, 'N/A')

        # Get isoforms and select the longest one for each junction
        isoforms1 = transcript_dict.get(gene_id, {}).get(junction1.split(':')[1], [])
        isoforms2 = transcript_dict.get(gene_id, {}).get(junction2.split(':')[1], [])
        longest_isoform1 = get_longest_isoform(gene_id,isoforms1, protein_dict)
        longest_isoform2 = get_longest_isoform(gene_id,isoforms2, protein_dict)
        # Extract coordinates
        coord1 = coord_dict.get(junction1, 'N/A')
        coord2 = coord_dict.get(junction2, 'N/A')

        # Add new annotated row
        annotations.append([
            gene_symbol, longest_isoform1, longest_isoform2, coord1, coord2
        ])

    # Create DataFrame with annotations
    annotated_df = pd.DataFrame(
        annotations, 
        columns=['Gene_Symbol', 'Isoform_1|Length', 'Isoform_2|Length', 'Coord_1', 'Coord_2']
    )

    # Merge with original stats DataFrame
    final_df = pd.concat([annotated_df, stats_df], axis=1)

    # Save annotated file
    output_file = os.path.join(output_folder, os.path.basename(file_path).replace('.txt', '-annotated.txt'))
    final_df.to_csv(output_file, sep='\t', index=False)

# 6. Annotate a stats file
def annotate_iso_stats_file(file_path, transcript_dict, gene_symbol_dict, protein_dict, output_folder):
    stats_df = pd.read_csv(file_path, sep='\t')

    annotations = []  # Store annotated rows

    for index, row in stats_df.iterrows():
        feature = row['Feature']

        gene_id, isoform = feature.split(':')  # Extract gene ID
        gene_symbol = gene_symbol_dict.get(gene_id, 'N/A')

        # Get isoforms and select the longest one for each junction
        junctions = transcript_dict[gene_id][isoform]
        try:
            prot_len, nmd_status, longest = protein_dict[gene_id,isoform]
        except:
            prot_len, nmd_status, longest = 0, 'NMD', 'UNK'

        # Add new annotated row
        annotations.append([gene_symbol, junctions, f'{prot_len}/{longest}', nmd_status])

    # Create DataFrame with annotations
    annotated_df = pd.DataFrame(
        annotations, 
        columns=['Gene_Symbol', 'Junctions', 'Protein Length', 'Predicted NMD']
    )

    # Merge with original stats DataFrame
    final_df = pd.concat([annotated_df, stats_df], axis=1)

    # Save annotated file
    output_file = os.path.join(output_folder, os.path.basename(file_path).replace('.txt', '-annotated.txt'))
    final_df.to_csv(output_file, sep='\t', index=False)

# 7. Annotate all stats files in the folder
def annotate_all_stats_files(stats_folder, transcript_dict, gene_symbol_dict, coord_dict, protein_dict, dataType='junction'):
    output_folder = stats_folder

    # Get a list of all .txt files excluding '-annotated.txt' files
    stats_files = [f for f in glob.glob(f"{stats_folder}/*.txt") if '-annotated' not in f]

    # Use tqdm to show progress over the stats files
    for stats_file in tqdm(stats_files, desc="Annotating Stats Files"):
        # Check if the annotated version already exists
        annotated_file = os.path.join(output_folder, os.path.basename(stats_file).replace('.txt', '-annotated.txt'))
        if os.path.exists(annotated_file):
            #print(f"Skipping {stats_file}, already annotated.")
            continue  # Skip this file

        try:
            if dataType == 'junction':
                annotate_junction_stats_file(stats_file, transcript_dict, gene_symbol_dict, coord_dict, protein_dict, output_folder)
            else:
                annotate_iso_stats_file(stats_file, transcript_dict, gene_symbol_dict, protein_dict, output_folder)
        except ZeroDivisionError as e:
            print(f"Error processing {stats_file}: {e}")
            continue


def annotate(gene_symbol_file,transcript_assoc_file,junction_coords_file,protein_summary_file,stats_folder,dataType='junction'):
    if junction_coords_file == None:
        #isoform expression/ratio
        coord_dict = None
    else:
        coord_dict = load_junction_coordinates(junction_coords_file)
    transcript_dict = load_transcript_associations(transcript_assoc_file,dataType=dataType)
    gene_symbol_dict = load_gene_symbols(gene_symbol_file)
    protein_dict = load_protein_summary(protein_summary_file)

    # Annotate all stats files
    annotate_all_stats_files(stats_folder, transcript_dict, gene_symbol_dict, coord_dict, protein_dict,dataType=dataType)




import pytest
import pathlib
from altanalyze3.utilities.io import (
    get_all_bam_chr,
    get_all_ref_chr,
    is_bam_paired,
    is_bam_indexed,
    get_indexed_references
)


DATA_FOLDER = pathlib.Path(__file__).resolve().parents[1].joinpath("data")

@pytest.mark.parametrize(
    "location, control_ref_chr",
    [
        (
            "gene_model_all.tsv",
            [
                "chr1", "chr10", "chr11", "chr12", "chr13",
                "chr14", "chr15", "chr16", "chr17", "chr18",
                "chr19", "chr2", "chr20", "chr21", "chr22",
                "chr3", "chr4", "chr5", "chr6", "chr7", "chr8", "chr9",
                "chrMT", "chrX", "chrY", "chrKI270750.1",
                "chrGL000009.2", "chrGL000194.1", "chrGL000195.1",
                "chrGL000205.2", "chrGL000213.1", "chrGL000216.2",
                "chrGL000218.1", "chrGL000219.1", "chrGL000220.1",
                "chrGL000225.1", "chrKI270442.1", "chrKI270711.1",
                "chrKI270713.1", "chrKI270721.1", "chrKI270726.1",
                "chrKI270727.1", "chrKI270728.1", "chrKI270731.1",
                "chrKI270733.1", "chrKI270734.1", "chrKI270744.1",
            ]
        )
    ]
)
def test_get_all_ref_chr(tmp_path, location, control_ref_chr):                           # this will also check guard_chr function
    calculated_ref_chr = get_all_ref_chr(
        get_indexed_references(
            DATA_FOLDER.joinpath(location),
            tmp_path                                                                     # https://docs.pytest.org/en/7.1.x/how-to/tmp_path.html#tmp-path
        ),
        1
    )
    assert sorted(calculated_ref_chr) == sorted(control_ref_chr)


@pytest.mark.parametrize(
    "location, control_bam_chr",
    [
        (
            "Cal27P5-1.bam",
            [
                "chr1", "chr10", "chr11", "chr12", "chr13", "chr14",
                "chr15", "chr16", "chr17", "chr18", "chr19", "chr2",
                "chr20", "chr21", "chr22", "chr3", "chr4", "chr5",
                "chr6", "chr7", "chr8", "chr9", "chrX", "chrY"
            ]
        )
    ]
)
def test_get_all_bam_chr(location, control_bam_chr):                           # this will also check guard_chr function
    calculated_bam_chr = get_all_bam_chr(DATA_FOLDER.joinpath(location), 1)
    assert sorted(calculated_bam_chr) == sorted(control_bam_chr)


@pytest.mark.parametrize(
    "location, control_is_paired",
    [
        ("Cal27P5-1.bam", True)
    ]
)
def test_is_bam_paired(location, control_is_paired):                           # this will also check skip_bam_read function
    calculated_is_paired = is_bam_paired(DATA_FOLDER.joinpath(location), 1)
    assert calculated_is_paired == control_is_paired


@pytest.mark.parametrize(
    "location, control_is_indexed",
    [
        ("Cal27P5-1.bam", True)
    ]
)
def test_is_bam_indexed(location, control_is_indexed):
    calculated_is_indexed = is_bam_indexed(DATA_FOLDER.joinpath(location), 1)
    assert calculated_is_indexed == control_is_indexed


import pysam
import shutil
import logging
import multiprocessing
from functools import partial

from altanalyze3.utilities.logger import setup_logger
from altanalyze3.utilities.constants import (
    IntRetCat,
    IntRetRawData,
    Job
)
from altanalyze3.utilities.helpers import get_tmp_suffix
from altanalyze3.utilities.io import (
    get_all_bam_chr,
    get_all_ref_chr,
    get_correct_contig,
    skip_bam_read,
    is_bam_paired
)


class IntronOverlaps:
    """
    Class to store counters for each inron overlapped with at least one read.
    The instance of this class can be used as Iterator.
    """

    def __init__(self):
        self.reset()

    def reset(self):
        self.overlaps = {}

    def __getitem__(self, key):                                                # key is a tuple (contig, start, end, name, strand)
        return self.overlaps.setdefault(key, {"p5": 0, "p3": 0})

    def __iter__(self):
        for (contig, start, end, name, strand), counters in self.overlaps.items():
            yield contig, start, end, name, strand, counters["p5"], counters["p3"]

    def increment_p5(self, key, step=None):                                    # key is a tuple (contig, start, end, name, strand)
        step = 1 if step is None else step
        self[key]["p5"] += step
        logging.debug(f"""Increment p5 counter on {step} for {key}""")

    def increment_p3(self, key, step=None):                                    # key is a tuple (contig, start, end, name, strand)
        step = 1 if step is None else step
        self[key]["p3"] += step
        logging.debug(f"""Increment p3 counter on {step} for {key}""")


class Counter:

    def __init__(self, bam, ref, span, strandness, location, threads=None):
        self.bam = bam
        self.ref = ref
        self.span = span
        self.strandness = strandness
        self.location = location
        self.threads = 1 if threads is None else threads
        self.paired = is_bam_paired(self.bam, self.threads)
        self.reset()

    def reset(self):
        self.cache = {}
        self.overlaps = IntronOverlaps()
        self.used_reads = {
            IntRetCat.PRIME_5: [],
            IntRetCat.PRIME_3: [],
            IntRetCat.DISCARD: []
        }

    def get_overlap_category(self, raw_data, span=None):
        span = self.span if span is None else span
        if raw_data.read_end - raw_data.intron_start >= span and \
                raw_data.intron_start - raw_data.read_start >= span:
            return IntRetCat.PRIME_5
        elif raw_data.read_start - raw_data.intron_start >= 0 and \
                raw_data.intron_end - raw_data.read_end >= 0:
            return IntRetCat.INTRON
        elif raw_data.read_end - raw_data.intron_end >= span and \
                raw_data.intron_end - raw_data.read_start >= span:
            return IntRetCat.PRIME_3
        else:
            return IntRetCat.DISCARD

    def guard_strandness(function):
        def check(self, current_data, cached_data=None):
            if self.strandness is IntRetCat.AUTO:
                if self.paired and \
                    (
                        current_data.xs_strand == cached_data.xs_strand == None or                                                    # downgrade to "unstranded"
                        current_data.xs_strand == cached_data.xs_strand == current_data.intron_strand == cached_data.intron_strand    # pass strandness check
                    ):
                    return function(self, current_data, cached_data)
                elif not self.paired and \
                    (
                        current_data.xs_strand is None or                                                                             # downgrade to "unstranded"
                        current_data.xs_strand == current_data.intron_strand                                                          # pass strandness check
                    ):
                    return function(self, current_data)
                else:
                    logging.debug("Strandness guard blocked the overlap for")
                    logging.debug(f"""{current_data}""")
                    return IntRetCat.DISCARD
            elif self.strandness is IntRetCat.UNSTRANDED:
                if self.paired:
                    return function(self, current_data, cached_data)
                else:
                    return function(self, current_data)
            elif self.strandness is IntRetCat.FORWARD:
                if self.paired and current_data.intron_strand == "+" and \
                    current_data.read_1 and current_data.read_strand == "+" and \
                        cached_data.read_2 and cached_data.read_strand == "-":
                    return function(self, current_data, cached_data)
                elif not self.paired and current_data.intron_strand == "+" and \
                    current_data.read_strand == "+":                                                           # all of the reads came from the "+" strand
                    return function(self, current_data)
                else:
                    logging.debug("Strandness guard blocked the overlap for")
                    logging.debug(f"""{current_data}""")
                    return IntRetCat.DISCARD
            elif self.strandness is IntRetCat.REVERSE:
                if self.paired and current_data.intron_strand == "-" and \
                    current_data.read_1 and current_data.read_strand == "-" and \
                        cached_data.read_2 and cached_data.read_strand == "+":
                    return function(self, current_data, cached_data)
                elif not self.paired and current_data.intron_strand == "-" and \
                    current_data.read_strand == "-":                                                           # all of the reads came from the "-" strand
                    return function(self, current_data)
                else:
                    logging.debug("Strandness guard blocked the overlap for")
                    logging.debug(f"""{current_data}""")
                    return IntRetCat.DISCARD
        return check

    def guard_distance(function):
        def check(self, current_data, cached_data=None):
            if not self.paired:
                return function(self, current_data)
            elif self.paired and current_data[0:5] == cached_data[0:5]:
                return function(self, current_data, cached_data)
            else:
                logging.debug("Distance guard blocked the overlap for")
                logging.debug(f"""{current_data}""")
                logging.debug(f"""{cached_data}""")
                return IntRetCat.DISCARD
        return check

    @guard_distance
    @guard_strandness
    def update_overlaps(self, current_data, cached_data=None):
        current_category = self.get_overlap_category(current_data)
        intron_key = (                                                  # will be the same for both current_data and cached_data because of guard_distance
            current_data.contig,
            current_data.intron_start,
            current_data.intron_end,
            current_data.intron_name,
            current_data.intron_strand
        )
        if not self.paired:
            logging.debug("Check overlap for single read")
            logging.debug(f"""{current_data}, {current_category}""")
            if current_category is IntRetCat.PRIME_5:
                self.overlaps.increment_p5(intron_key)
            elif current_category is IntRetCat.PRIME_3:
                self.overlaps.increment_p3(intron_key)
            return current_category
        else:
            assert(cached_data != None)
            logging.debug("Check overlap for paired-end")
            cached_category = self.get_overlap_category(cached_data)
            logging.debug(f"""{current_data}, {current_category}""")
            logging.debug(f"""{cached_data}, {cached_category}""")
            if IntRetCat.DISCARD in [current_category, cached_category] or \
                    current_category == cached_category:                                                           # both introns, both 5', or both 3'
                return IntRetCat.DISCARD
            elif current_category in [IntRetCat.PRIME_5, IntRetCat.INTRON] and cached_category in [IntRetCat.PRIME_5, IntRetCat.INTRON]:   # 5' and intron or intron and 5'
                self.overlaps.increment_p5(intron_key)
                return IntRetCat.PRIME_5
            elif current_category in [IntRetCat.PRIME_3, IntRetCat.INTRON] and cached_category in [IntRetCat.PRIME_3, IntRetCat.INTRON]:   # 3' and intron or intron and 3'
                self.overlaps.increment_p3(intron_key)
                return IntRetCat.PRIME_3
            else:
                logging.debug(f"""Not implemented combination of {current_category} and {cached_category} categories""")
                return IntRetCat.DISCARD

    def calculate(self, contig):
        with pysam.AlignmentFile(self.bam, mode="rb", threads=self.threads) as bam_handler:
            with pysam.TabixFile(str(self.ref), mode="r", parser=pysam.asBed(), threads=self.threads) as ref_handler:
                contig_ref = get_correct_contig(contig, ref_handler)                                                # contig in the file can be both with or without 'chr' prefix
                contig_bam = get_correct_contig(contig, bam_handler)                                                # the same as above
                intron_iter = ref_handler.fetch(contig_ref)
                intron = next(intron_iter)                                                                          # get initial value from intron iterator
                no_introns = False                                                                                  # to break the outer fetching reads loop
                for read in bam_handler.fetch(contig_bam):                                                          # fetches only mapped reads
                    logging.debug(
                        f"""Fetch a read {read.query_name} {contig_bam}:{read.reference_start}-{read.reference_end} {"-" if read.is_reverse else "+"}"""
                    )
                    if skip_bam_read(read):                                                                        # gate to skip all "bad" reads
                        logging.debug(f"""Skip a read {read.query_name}""")
                        logging.debug(f"""is_secondary: {read.is_secondary}""")
                        logging.debug(f"""is_duplicate: {read.is_duplicate}""")
                        logging.debug(f"""is_supplementary: {read.is_supplementary}""")
                        logging.debug(f"""is_paired and mate_is_unmapped: {read.is_paired and read.mate_is_unmapped}""")
                        continue
                    while read.reference_start - intron.end >= 0:
                        try:
                            intron = next(intron_iter)
                            logging.debug(
                                f"""Switched to a new intron {intron.name} {contig_ref}:{intron.start}-{intron.end} {intron.strand}"""
                            )
                        except StopIteration:
                            no_introns = True
                            break
                    if no_introns:                                                                              # no need to iterate over the reads if no introns left
                        logging.debug("Halt read iteration - run out of introns")
                        break
                    xs_strand = None
                    try:
                        xs_strand = read.get_tag("XS")
                    except KeyError:
                        pass
                    current_data = IntRetRawData(
                        contig = contig,
                        intron_start = intron.start,
                        intron_end = intron.end,
                        intron_name = intron.name,
                        intron_strand = intron.strand,
                        read_start = read.reference_start,
                        read_end = read.reference_end,
                        read_strand = "-" if read.is_reverse else "+",
                        xs_strand = xs_strand,
                        read_name = read.query_name,
                        read_1 = read.is_read1,
                        read_2 = read.is_read2,
                    )
                    if self.paired:
                        if read.query_name in self.cache:
                            cached_data = self.cache[read.query_name]
                            overlapped_as = self.update_overlaps(current_data, cached_data)
                            self.used_reads[overlapped_as].append(read.query_name)
                            del self.cache[read.query_name]
                            logging.debug(f"""Remove cached data for {read.query_name}""")
                        else:
                            self.cache[read.query_name] = current_data
                            logging.debug(f"""Add cached data for {read.query_name}""")
                    else:
                        overlapped_as = self.update_overlaps(current_data)
                        self.used_reads[overlapped_as].append(read.query_name)

    def export_counts(self):
        logging.info(f"""Save counts to {self.location}""")
        with self.location.open("w") as out_handler:
            for contig, start, end, name, strand, p5, p3 in self.overlaps:
                out_handler.write(f"{contig}\t{start-self.span-1}\t{start+self.span-1}\t{name}_{start}\t{p5}\t{strand}\n")
                out_handler.write(f"{contig}\t{end-self.span}\t{end+self.span}\t{name}_{end}\t{p3}\t{strand}\n")

    def export_reads(self):
        bam_location = self.location.with_suffix(".bam")
        logging.info(f"""Save reads to {bam_location}""")
        with pysam.AlignmentFile(self.bam, mode="rb", threads=self.threads) as in_bam_handler:
            with pysam.AlignmentFile(bam_location, "wb", threads=self.threads, template=in_bam_handler) as out_bam_handler:
                for read in in_bam_handler.fetch():
                    logging.debug(f"""Fetch read {read.query_name}""")
                    if skip_bam_read(read):
                        logging.debug("Skip")
                        continue
                    if read.query_name in self.used_reads[IntRetCat.PRIME_5]:
                        read.set_tag("XI", "P5")
                        logging.debug("Assign XI=P5")
                    elif read.query_name in self.used_reads[IntRetCat.PRIME_3]:
                        read.set_tag("XI", "P3")
                        logging.debug("Assign XI=P3")
                    elif read.query_name in self.used_reads[IntRetCat.DISCARD]:
                        read.set_tag("XI", "D")
                        logging.debug("Assign XI=D")
                    else:
                        read.set_tag("XI", "U")
                        logging.debug("Assign XI=U")
                    out_bam_handler.write(read)


def get_jobs(args):
    return [
        Job(
            contig=contig,                                                            # contig is always prepended with 'chr'
            location=args.tmp.joinpath(get_tmp_suffix())
        )
        for contig in get_all_bam_chr(args.bam, args.threads)
            if contig in get_all_ref_chr(args.ref, args.threads) and contig in args.chr                 # safety measure to include only chromosomes present in BAM, BED, and --chr
    ]


def process_contig(args, job):
    setup_logger(
        multiprocessing.get_logger(),
        args.loglevel
    )
    multiprocessing.current_process().name = job.contig
    logging.info(f"""Process chromosome {job.contig} to {job.location}""")
    counter = Counter(
        bam=args.bam,
        ref=args.ref,
        span=args.span,
        strandness=args.strandness,
        location=job.location,
        threads=args.threads
    )
    counter.calculate(job[0])
    counter.export_counts()
    if args.savereads:
        counter.export_reads()


def collect_results(args, jobs):
    with args.output.with_suffix(".bed").open("w") as output_stream:
        for job in jobs:
            logging.info(f"""Collect counts from {job.location}""")
            with job.location.open("r") as input_stream:
                output_stream.write(input_stream.read())
                logging.debug(f"""Remove {job.location}""")
                job.location.unlink()
    if args.savereads:
        tmp_bam = args.output.with_suffix(get_tmp_suffix() + ".bam")
        with pysam.AlignmentFile(args.bam, mode="rb", threads=args.threads) as template_handler:
            with pysam.AlignmentFile(tmp_bam, "wb", threads=args.threads, template=template_handler) as out_bam_handler:
                for job in jobs:
                    bam_location = job.location.with_suffix(".bam")
                    logging.info(f"""Collect reads from {bam_location}""")
                    with pysam.AlignmentFile(bam_location, mode="rb", threads=args.threads) as in_bam_handler:
                        for read in in_bam_handler.fetch(until_eof=True):                                         # we don't need index because of until_eof
                            out_bam_handler.write(read)
                        logging.debug(f"""Remove {bam_location}""")
                        bam_location.unlink()
        pysam.sort("-o", str(args.output.with_suffix(".bam")), str(tmp_bam))
        pysam.index(str(args.output.with_suffix(".bam")))
        tmp_bam.unlink()


def count_introns(args):
    jobs = get_jobs(args)
    logging.info(f"""Span {len(jobs)} job(s) between {args.cpus} CPU(s)""")
    with multiprocessing.Pool(args.cpus) as pool:
        pool.map(partial(process_contig, args), jobs)
    collect_results(args, jobs)

    logging.debug("Removing temporary directory")
    shutil.rmtree(args.tmp)


import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from scipy.stats import zscore
from matplotlib.colors import ListedColormap
from matplotlib import cm
from matplotlib.colors import to_hex

# Define a colormap to use for marker expression heatmap
N = 256
vals = np.ones((N*2, 4))
vals[:N, 0] = np.linspace(15/256, 0, N)
vals[:N, 1] = np.linspace(255/256, 0, N)
vals[:N, 2] = np.linspace(255/256, 0, N)
vals[N:, 0] = np.linspace(0, 255/256, N)
vals[N:, 1] = np.linspace(0, 243/256, N)
vals[N:, 2] = np.linspace(0, 15/256, N)
blue_black_yellow_cmap = ListedColormap(vals)


def assign_rainbow_colors_to_groups(groups):
    """ Creates a dictionary of cluster names to hexadecimal color strings

    This function takes a list of groups and assigns each unique item in the
    groups a color (using the matplotlib.cm.rainbow color-map) as a hexadecimal
    string value. This is useful for storing a single color scheme for clusters
    to be used with downstream visualizations.

    Arguments
    ---------
    groups : numpy.Array[str]
        List of cluster names. (Required)

    Returns
    -------
    dict {str:str}
        Dictionary of cluster-names to assigned colors (hexadecimal value)

    """
    unique_groups = np.unique(groups)
    groups_to_num = pd.Series(list(range(len(unique_groups))),
                              index=unique_groups)
    n = len(unique_groups)
    groups_to_color = pd.Series([to_hex(cm.rainbow(item / n)) for item in
                                 groups_to_num.values], index=groups_to_num.index.values).to_dict()
    return groups_to_color


def plot_markers_df(marker_heatmap, markers_df, clusters, path_to_save_figure):
    """ Plots a heatmap of the MarkerFinder results
    Arguments
    ---------
    marker_heatmap : pandas.DataFrame
        Data to plot. The rows must intersect with features in the \
        ordered_markers_df and the columns must intersect with the cells in \
        clusters.
    markers_df : pandas.DataFrame
        The markers_df where "marker" column are the genes and the "top_cluster" is the cluster to which the gene belongs to  (Required)
    clusters : pandas.DataFrame
        The cluster assignments of the samples, index of this data frame are the sample names (Required)
    path_to_save_figure : str
        The path to save the figure. (Required)

    Returns
    -------
    None
        Saves the heatmap to the file specified by path_to_save_figure

    """
    try:
        # get Dictionary of cluster-names to assigned colors (hexadecimal value)
        groups_to_colors = assign_rainbow_colors_to_groups(groups=np.array(clusters["cluster"]))

        # Filter the input data matrix to the markers and cells of interest
        input_df = marker_heatmap.loc[markers_df["marker"].values, clusters.index]
        input_df = input_df.astype(float)

        # Map the order of the groups_to_colors to make a cmap
        groups_to_order = pd.Series(list(range(len(groups_to_colors))),
                                    index=groups_to_colors.keys())

        # Build top heatmap to label clusters
        cell_labels_df = pd.DataFrame({"cluster": [groups_to_order[item] for item in clusters["cluster"].values]},
                                      index=clusters.index).T

        # Build row annotation dataframe
        row_labels_df = pd.DataFrame({"cluster": [groups_to_order[item] for item in markers_df["top_cluster"].values]},
                                     index=markers_df["marker"].values)

        # Get top markers for each cluster
        top_markers = markers_df.groupby('top_cluster')['marker'].first()

        # Build df to add cluster label ticks
        label_to_position = pd.pivot_table(pd.DataFrame({
            "cluster": clusters["cluster"].values,
            "position": list(range(clusters.shape[0]))}),
            index="cluster",
            values="position",
            aggfunc=np.mean)["position"].sort_values()

        # Build df to add marker label ticks
        label_to_position2 = pd.pivot_table(pd.DataFrame({
            "cluster": markers_df["top_cluster"].values,
            "position": list(range(markers_df.shape[0]))}),
            index="cluster",
            values="position",
            aggfunc=np.mean)["position"].sort_values()

        # Z-score normalize the expression matrix
        z_df = input_df.apply(lambda x: pd.Series(zscore(x.values),
                                                  index=x.index.values), axis=1)

        # Build the heatmap
        plt.close("all")
        fig = plt.figure(constrained_layout=True, figsize=(9, 5))
        ax = fig.add_gridspec(2, 2, width_ratios=(1, 20), height_ratios=(1, 20),
                      wspace=0.005, hspace=0.01)  # Adjusted to create space for row annotation and top markers
        ax2 = fig.add_subplot(ax[1, 1])
        ax1 = fig.add_subplot(ax[0, 1])
        ax3 = fig.add_subplot(ax[1, 0])  # Subplot for row annotation

        heat1 = sns.heatmap(cell_labels_df,
                            yticklabels=False,
                            xticklabels=False,
                            cmap=sns.color_palette(groups_to_colors.values()),
                            cbar=False,
                            ax=ax1)
        ax1.set_xticks(label_to_position.values)
        ax1.set_xticklabels(label_to_position.index.values, rotation=45, ha="left")
        ax1.xaxis.tick_top()

        heat2 = sns.heatmap(z_df,
                            vmin=-3,
                            vmax=3,
                            cmap=blue_black_yellow_cmap,  # Assuming you have blue_black_yellow_cmap defined
                            xticklabels=False,
                            yticklabels=False,  # Disable yticklabels here
                            cbar=True,
                            cbar_kws={"shrink": 0.5},
                            ax=ax2)
        ax2.collections[0].colorbar.set_label("Z-Score Normalized Expression")

        heat3 = sns.heatmap(row_labels_df,
                            xticklabels=False,
                            yticklabels=False,  # Disable yticklabels here as well
                            cmap=sns.color_palette(groups_to_colors.values()),
                            cbar=False,
                            ax=ax3)
        ax3.set_yticks(label_to_position2.values)
        ax3.set_yticklabels(label_to_position.index.values)
        # ax3.yaxis.tick_left()
        fig.suptitle('Marker Finder Heatmap', fontsize=16)
        # fig.tight_layout()
        plt.savefig(path_to_save_figure, dpi=10, bbox_inches='tight', format="pdf", pad_inches=0.9)

    except Exception as e:
        print(str(e))
        print("Warning! Failed to run plot_markers_df. See above Exception.")

# Usage example (this part should be replaced with actual data in your script)
# marker_heatmap = pd.DataFrame(...)
# markers_df = pd.DataFrame(...)
# clusters = pd.DataFrame(...)
# groups_to_colors = {'cluster1': '#1f77b4', 'cluster2': '#ff7f0e', ...}
# path_to_save_figure = 'heatmap.png'
# plot_markers_df(marker_heatmap, markers_df, clusters, groups_to_colors, path_to_save_figure)


def plot_metadata_heatmap(udon_clusters, udon_metadata, metadata_col, path_to_save_figure):
    try:
        # get Dictionary of cluster-names to assigned colors (hexadecimal value)
        groups_to_colors = assign_rainbow_colors_to_groups(groups=np.array(udon_clusters["cluster"]))

        # Map the order of the groups_to_colors to make a cmap
        groups_to_order = pd.Series(list(range(len(groups_to_colors))),
                                    index=groups_to_colors.keys())

        # Build top heatmap to label clusters
        cell_labels_df = pd.DataFrame({"cluster": [groups_to_order[item] for item in udon_clusters["cluster"].values]},
                                      index=udon_clusters.index).T


        # Build df to add cluster label ticks
        label_to_position = pd.pivot_table(pd.DataFrame({
            "cluster": udon_clusters["cluster"].values,
            "position": list(range(udon_clusters.shape[0]))}),
            index="cluster",
            values="position",
            aggfunc=np.mean)["position"].sort_values()

        # Get unique clusters and samples
        unique_clusters = udon_metadata[metadata_col].unique()
        samples = udon_metadata.index

        # Initialize the binary matrix with zeros
        binary_matrix = pd.DataFrame(0, index=unique_clusters, columns=samples)

        # Populate the binary matrix
        for sample in samples:
            cluster = udon_metadata.loc[sample, metadata_col]
            binary_matrix.loc[cluster, sample] = 1

        binary_matrix = binary_matrix.sort_index()
        print("sorted index")

        # Build df to add cluster label ticks
        label_to_position2 = pd.pivot_table(pd.DataFrame({
            "cluster": binary_matrix.index,
            "position": list(range(binary_matrix.shape[0]))}),
            index="cluster",
            values="position",
            aggfunc=np.mean)["position"].sort_values()

        # Build the heatmap
        plt.close("all")
        fig = plt.figure(constrained_layout=True, figsize=(9, 6))
        ax = fig.add_gridspec(2, 1, height_ratios=(1, 20),
                                  wspace=0.005,
                                  hspace=0.01)  # Adjusted to create space for row annotation and top markers
        ax2 = fig.add_subplot(ax[1, 0])
        ax1 = fig.add_subplot(ax[0, 0])

        heat1 = sns.heatmap(cell_labels_df,
                                yticklabels=False,
                                xticklabels=False,
                                cmap=sns.color_palette(groups_to_colors.values()),
                                cbar=False,
                                ax=ax1)
        ax1.set_xticks(label_to_position.values)
        ax1.set_xticklabels(label_to_position.index.values, rotation=45, ha="left")
        ax1.xaxis.tick_top()

        heat2 = sns.heatmap(binary_matrix,
                                vmin=0,
                                vmax=1,
                                cmap='YlGn',
                                xticklabels=False,
                                yticklabels=False,  # Disable yticklabels here
                                cbar=True,
                                cbar_kws={"shrink": 0.5},
                                ax=ax2)
        ax2.collections[0].colorbar.set_label("Z-Score Normalized Expression")
        ax2.set_yticks(label_to_position2.values)
        ax2.set_yticklabels(label_to_position2.index.values, ha="right")
        ax2.yaxis.tick_left()

        fig.suptitle('Metadata Heatmap', fontsize=16)
        # fig.tight_layout()
        plt.savefig(path_to_save_figure, dpi=10, bbox_inches='tight', format="pdf", pad_inches=0.9)

    except Exception as e:
        print(str(e))
        print("Warning! Failed to run plot_metadata_heatmap. See above Exception.")


def plot_markers_df_subplots(marker_heatmap, markers_df, clusters, groups_to_colors, path_to_save_figure):
    """
    Plots a heatmap of the MarkerFinder results
    Arguments
    ---------
    marker_heatmap : pandas.DataFrame
        Data to plot. The rows must intersect with features in the \
        ordered_markers_df and the columns must intersect with the cells in clusters.
    markers_df : pandas.DataFrame
        The markers_df where "marker" column are the genes and the "top_cluster" is the cluster to which the gene belongs to  (Required)
    clusters : pandas.DataFrame
        The cluster assignments of the samples, index of this data frame are the sample names (Required)
    groups_to_colors : dict {str:str}
        Dictionary of cluster-names to assigned colors (hexadecimal value) \
        The pyInfinityFlow.Plotting_Utilities.assign_rainbow_colors_to_groups \
        can be used to generate this dictionary from a list of clusters. \
        (Required)
    path_to_save_figure : str
        The path to save the figure. (Required)

    Returns
    -------
    None
        Saves the heatmap to the file specified by path_to_save_figure
    """
    try:
        # Filter the input data matrix to the markers and cells of interest
        input_df = marker_heatmap.loc[markers_df["marker"].values, clusters.index]
        input_df = input_df.astype(float)

        # Map the order of the groups_to_colors to make a cmap
        groups_to_order = pd.Series(list(range(len(groups_to_colors))),
                                    index=groups_to_colors.keys())

        # Build top heatmap to label clusters
        cell_labels_df = pd.DataFrame({"cluster": [groups_to_order[item] for item in clusters["cluster"].values]},
                                      index=clusters.index).T

        # Build row annotation dataframe
        row_labels_df = pd.DataFrame({"cluster": [groups_to_order[item] for item in markers_df["top_cluster"].values]},
                                     index=markers_df["marker"].values)

        # Get top markers for each cluster
        top_markers = markers_df.groupby('top_cluster')['marker'].first()

        # Build df to add cluster label ticks
        label_to_position = pd.pivot_table(pd.DataFrame({
            "cluster": clusters["cluster"].values,
            "position": list(range(clusters.shape[0]))}),
            index="cluster",
            values="position",
            aggfunc=np.mean)["position"].sort_values()

        # Build df to add marker label ticks
        label_to_position2 = pd.pivot_table(pd.DataFrame({
            "cluster": markers_df["top_cluster"].values,
            "position": list(range(markers_df.shape[0]))}),
            index="cluster",
            values="position",
            aggfunc=np.mean)["position"].sort_values()

        # Z-score normalize the expression matrix
        z_df = input_df.apply(lambda x: pd.Series(zscore(x.values),
                                                  index=x.index.values), axis=1)

        # Build the heatmap
        plt.close("all")
        fig, axs = plt.subplots(2, 2, figsize=(18, 10),
                                gridspec_kw={'width_ratios': [1, 20], 'height_ratios': [1, 20], 'wspace': 0.005,
                                             'hspace': 0.01})
        ax1, ax2, ax3 = axs[0, 1], axs[1, 1], axs[1, 0]  # ax3 remains the same

        heat1 = sns.heatmap(cell_labels_df,
                            yticklabels=False,
                            xticklabels=False,
                            cmap=sns.color_palette(groups_to_colors.values()),
                            cbar=False,
                            ax=ax1)
        ax1.set_xticks(label_to_position.values)
        ax1.set_xticklabels(label_to_position.index.values, rotation=45, ha="left")
        ax1.xaxis.tick_top()

        heat2 = sns.heatmap(z_df,
                            vmin=-3,
                            vmax=3,
                            cmap=blue_black_yellow_cmap,  # Assuming you have blue_black_yellow_cmap defined
                            xticklabels=False,
                            yticklabels=False,  # Disable yticklabels here
                            cbar=True,
                            cbar_kws={"shrink": 0.5},
                            ax=ax2)
        ax2.collections[0].colorbar.set_label("Z-Score Normalized Expression")

        heat3 = sns.heatmap(row_labels_df,
                            xticklabels=False,
                            yticklabels=False,  # Disable yticklabels here as well
                            cmap=sns.color_palette(groups_to_colors.values()),
                            cbar=False,
                            ax=ax3)
        ax3.set_yticks(label_to_position2.values)
        ax3.set_yticklabels(label_to_position.index.values)

        fig.suptitle('Marker Finder Heatmap', fontsize=16)

        # Save the figure with compression
        plt.savefig(path_to_save_figure, dpi=100, bbox_inches='tight', format="pdf", pad_inches=0.9)

    except Exception as e:
        print(str(e))
        print("Warning! Failed to run plot_markers_df. See above Exception.")



from setuptools import setup, find_packages
import os

def parse_requirements():
    with open('requirements.txt') as f:
        return f.read().splitlines()

setup(
    name="altanalyze3",
    version="0.1.1",
    description="AltAnalyze3",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="AltAnalyze Development Team",
    author_email="altanalyze@gmail.com",
    url="https://github.com/SalomonisLab/altanalyze3.git",
    license="Apache-2.0",
    include_package_data=True,
    packages=find_packages(exclude=["docs", "tests", "cwls", "tmp"]),
    install_requires=parse_requirements(),
    scripts=["altanalyze3/bin/altanalyze3"],
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Scientific/Engineering :: Bio-Informatics"
    ],
    zip_safe=False,
)
